# -*- coding: utf-8 -*-
# File: config.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Config functions for setting up Tensorpack Faster-RCNN models and training schemes.


Description of the config file.

Most of the descriptions are taken from

<https://github.com/tensorpack/tensorpack/blob/master/examples/FasterRCNN/config.py>

Backbone settings


**BACKBONE**

.BOTTLENECK: Resnet oder resnext_32xd4

.FREEZE_AFFINE: Do not train affine parameters inside norm layers

.FREEZE_AT: Options: 0, 1, 2. How many stages in backbone to freeze (not training)

.NORM: options: FreezeBN, SyncBN, GN, None

.RESNET_NUM_BLOCKS: For resnet50: [3,4,6,3], for resnet101: [3,4,23,3]

.TF_PAD_MODE: Use a base model with TF-preferred padding mode, which may pad more pixels on right or bottom
than top/left. See https://github.com/tensorflow/tensorflow/issues/18213. Using either one should probably give the same
performance.


**CASCADE**

.BBOX_REG_WEIGHTS: Bounding box regression weights

.IOUS: Iou levels


**DATA**

.TRAIN_NUM_WORKERS: Number of threads to use when parallelizing the pre-processing (e.g. augmenting, adding anchors,
RPN gt-labelling,...)


**FPN**

.ANCHOR_STRIDES: Strides for each FPN level. Must be the same length as ANCHOR_SIZES

.CASCADE: Use Cascade RCNN

.FRCNN_CONV_HEAD_DIM: Dimension of conv head in FRCNN head

.FRCNN_FC_HEAD_DIM: Dimension(s) of fc layer in FRCNN head

.FRCNN_HEAD_FUNC: Stack of FRCNN head. Choices: fastrcnn_2fc_head, fastrcnn_4conv1fc_{,gn_}head

.MRCNN_HEAD_FUNC: Stack of MRCNN head. Choices: maskrcnn_up4conv_{,gn_}head

.NORM: Choices: 'None', 'GN'

.NUM_CHANNEL: Number of channels

.PROPOSAL_MODE: Choices: 'Level', 'Joint'

**FRCNN**

.BATCH_PER_IM: Number of total proposals selected. Will divide into fg and bg by given ratio

.BBOX_REG_WEIGHTS: Bounding box regression weights

.FG_RATIO: Fg ratio for proposals selection

.FG_THRESH: Threshold how to divide fg and bg selection

.MODE_MASK: Whether to train mask head

**MRCNN**

.ACCURATE_PASTE: Slightly more aligned results, but very slow on numpy

.HEAD_DIM: Head dimension

**PREPROC**

.MAX_SIZE: Maximum edge size

.PIXEL_MEAN: Pixel mean (on the training data set)

.PIXEL_STD: Pixel std (on the training data set)

.SHORT_EDGE_SIZE: Size to resize the image to (inference), while not exceeding max size

.TRAIN_SHORT_EDGE_SIZE: The size to resize the image to (training), while not exceeding max size. [min, max] to sample
from

**RPN**

.ANCHOR_RATIOS: Anchor ratios

.ANCHOR_SIZES: Anchor sizes

.ANCHOR_STRIDE: Anchor stride

.BATCH_PER_IM: Total (across FPN levels) number of anchors that are marked valid

.CROWD_OVERLAP_THRESH: Anchors which overlap with a crowd box (IOA larger than threshold) will be ignored. Setting this
to a value larger than 1.0 will disable the feature. It is disabled by default because Detectron does not do this.

.FG_RATIO: Fg ratio among selected RPN anchors

.HEAD_DIM: Deprecated

.MIN_SIZE: Minimal size length for proposals

.NEGATIVE_ANCHOR_THRESH: Negative anchor threshold

.POSITIVE_ANCHOR_THRESH: Positive anchor threshold

.PER_LEVEL_NMS_TOPK: Number of top k proposals after carrying out nms (inference). selection per level proposals

.TRAIN_PER_LEVEL_NMS_TOPK: Number of proposals after carrying out nms (training)

.TRAIN_PRE_NMS_TOPK: Number of top k proposals before carrying out nms (training)

.PRE_NMS_TOPK: Number of top k proposals before carrying out nms (inference)

.TRAIN_POST_NMS_TOPK: Number of proposals after carrying out nms (training)

.POST_NMS_TOPK: Number of proposals after carrying out nms (inference)

**OUTPUT**

.FRCNN_NMS_THRESH: Nms threshold for output. nms being performed per class prediction

.RESULTS_PER_IM: Number of output detection results

.RESULT_SCORE_THRESH: Threshold for detection result

.NMS_THRESH_CLASS_AGNOSTIC: Nms threshold for output. nms being performed over all class predictions

TRAINER: options: 'horovod', 'replicated'. Note that Horovod trainer is not available when TF2 is installed

**TRAIN**

.LR_SCHEDULE: "1x" schedule in detectron.  LR_SCHEDULE means equivalent steps when the total batch size is 8.
               It can be either a string like "3x" that refers to standard convention, or a list of int.
               LR_SCHEDULE=3x is the same as LR_SCHEDULE=[420000, 500000, 540000], which
               means to decrease LR at steps 420k and 500k and stop training at 540k.
               When the total bs!=8, the actual iterations to decrease learning rate, and
               the base learning rate are computed from BASE_LR and LR_SCHEDULE.
               Therefore, there is *no need* to modify the config if you only change the number of GPUs.

.EVAL_PERIOD: Will call eval callback every eval period

.CHECKPOINT_PERIOD: Will save model weights every checkpoint period

.WEIGHT_DECAY: Regularization weight decay

.BASE_LR: Base learning rate

.WARMUP: In terms of iterations. This is not affected by #GPUs

.WARMUP_INIT_LR: Defined for total batch size=8. Otherwise it will be adjusted automatically

.STEPS_PER_EPOCH: Steps per epoch. One steps is equivalent to the forward/backward path of one image.

.STARTING_EPOCH: Starting epoch. Useful when restarting training.

.LOG_DIR: Log dir

"""

import os
from typing import List, Mapping, Tuple

import numpy as np
from tensorpack.tfutils import collect_env_info  # pylint: disable=E0401
from tensorpack.utils import logger  # pylint: disable=E0401

# pylint: disable=import-error
from tensorpack.utils.gpu import get_num_gpu

from .....utils.metacfg import AttrDict
from .....utils.settings import ObjectTypes

# pylint: enable=import-error


__all__ = ["train_frcnn_config", "model_frcnn_config"]


def model_frcnn_config(config: AttrDict, categories: Mapping[str, ObjectTypes], print_summary: bool = True) -> None:
    """
    Sanity checks for Tensorpack Faster-RCNN config settings, where the focus lies on the model for predicting.
    It will update the config instance.

    :param config: Configuration instance as an AttrDict
    :param categories: Dict with category ids (int) and category names as keys and values.
    :param print_summary: Will optionally print the summary, the full learning rate schedule and the number of steps.
    """

    config.freeze(False)

    categories = {str(key): categories[val] for key, val in enumerate(categories, 1)}
    categories[0] = "BG"
    config.DATA.CLASS_NAMES = list(categories.values())
    config.DATA.CLASS_DICT = categories
    config.DATA.NUM_CATEGORY = len(config.DATA.CLASS_NAMES) - 1

    assert config.BACKBONE.BOTTLENECK in ["resnet", "resnext_32xd4"], config.BACKBONE.BOTTLENECK
    assert config.BACKBONE.NORM in ["FreezeBN", "SyncBN", "GN", "None"], config.BACKBONE.NORM
    if config.BACKBONE.NORM != "FreezeBN":
        assert not config.BACKBONE.FREEZE_AFFINE
    assert config.BACKBONE.FREEZE_AT in [0, 1, 2]

    config.RPN.NUM_ANCHOR = len(config.RPN.ANCHOR_SIZES) * len(config.RPN.ANCHOR_RATIOS)
    assert len(config.FPN.ANCHOR_STRIDES) == len(config.RPN.ANCHOR_SIZES)
    config.FPN.RESOLUTION_REQUIREMENT = config.FPN.ANCHOR_STRIDES[3]

    size_mult = config.FPN.RESOLUTION_REQUIREMENT * 1.0
    config.PREPROC.MAX_SIZE = np.ceil(config.PREPROC.MAX_SIZE / size_mult) * size_mult
    assert config.FPN.PROPOSAL_MODE in ["Level", "Joint"]
    assert config.FPN.FRCNN_HEAD_FUNC.endswith("_head")
    if config.MODE_MASK:
        assert config.FPN.MRCNN_HEAD_FUNC.endswith("_head")
    assert config.FPN.NORM in ["None", "GN"]

    if config.FPN.CASCADE:
        # the first threshold is the proposal sampling threshold
        assert config.CASCADE.IOUS[0] == config.FRCNN.FG_THRESH
        assert len(config.CASCADE.BBOX_REG_WEIGHTS) == len(config.CASCADE.IOUS)

    os.environ["TF_CUDNN_USE_AUTOTUNE"] = "0"
    config.NUM_GPUS = get_num_gpu()

    config.freeze()
    if print_summary:
        logger.info("Config: ------------------------------------------\n %s", config)


def train_frcnn_config(config: AttrDict) -> Tuple[List[Tuple[int, int]], List[Tuple[int, int]], int]:
    """
    Enhances the config instance by some parameters which are necessary for setting up the training
    Run some sanity checks, and populate some configs from others

    :param config: The model configuration
    :return: A tuple with warmup learning rate schedule
    """

    config.freeze(False)

    train_scales = config.PREPROC.TRAIN_SHORT_EDGE_SIZE
    if isinstance(train_scales, (list, tuple)) and train_scales[1] - train_scales[0] > 100:
        # don't autotune if augmentation is on
        os.environ["TF_CUDNN_USE_AUTOTUNE"] = "0"
    os.environ["TF_AUTOTUNE_THRESHOLD"] = "1"
    assert config.TRAINER in ["replicated"], config.TRAINER

    # setup NUM_GPUS
    assert "OMPI_COMM_WORLD_SIZE" not in os.environ
    logger.set_logger_dir(config.TRAIN.LOG_DIR, "d")
    number_gpu = get_num_gpu()
    logger.info("Environment Information:\n %s", collect_env_info())

    assert number_gpu > 0, "Has to train with GPU!"
    assert (
        number_gpu % 8 == 0 or 8 % number_gpu == 0
    ), f"Can only train with 1,2,4 or >=8 GPUs, but found {number_gpu} GPUs"

    learning_rate = config.TRAIN.LR_SCHEDULE
    if isinstance(learning_rate, str):
        if learning_rate.endswith("x"):
            lr_schedule_k_iter = {f"{k}x": [180 * k - 120, 180 * k - 40, 180 * k] for k in range(2, 10)}
            lr_schedule_k_iter["1x"] = [120, 160, 180]
            config.TRAIN.LR_SCHEDULE = [x * 1000 for x in lr_schedule_k_iter[learning_rate]]
        else:
            config.TRAIN.LR_SCHEDULE = eval(learning_rate)  # pylint: disable=W0123

    config.TRAIN.NUM_GPUS = number_gpu
    # Convert some config, that might be stored as strings
    config.TRAIN.BASE_LR = float(config.TRAIN.BASE_LR)
    config.TRAIN.WARMUP_INIT_LR = float(config.TRAIN.WARMUP_INIT_LR)
    config.TRAIN.WEIGHT_DECAY = float(config.TRAIN.WEIGHT_DECAY)
    # Compute the training schedule from the number of GPUs ...
    step_num = config.TRAIN.STEPS_PER_EPOCH
    # warmup is step based, lr is epoch based
    init_lr = config.TRAIN.WARMUP_INIT_LR * min(8.0 / config.TRAIN.NUM_GPUS, 1.0)
    warmup_schedule = [(0, init_lr), (config.TRAIN.WARMUP, config.TRAIN.BASE_LR)]
    warmup_end_epoch = config.TRAIN.WARMUP * 1.0 / step_num
    lr_schedule = [(int(warmup_end_epoch + 0.5), config.TRAIN.BASE_LR)]

    factor = 8.0 / config.TRAIN.NUM_GPUS
    for idx, steps in enumerate(config.TRAIN.LR_SCHEDULE[:-1]):
        mult = 0.1 ** (idx + 1)
        lr_schedule.append((steps * float(factor) // step_num, config.TRAIN.BASE_LR * float(mult)))

    config.freeze()
    logger.info("Config: \n %s", str(config), config.to_dict())
    return warmup_schedule, lr_schedule, step_num



# -*- coding: utf-8 -*-
# File: __init__.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Wrappers for models of external libraries as well as implementation of the Cascade-RCNN model of Tensorpack.
"""

from ..utils.file_utils import tensorpack_available
from .base import *
from .d2detect import *
from .deskew import *
from .doctrocr import *
from .fastlang import *
from .hfdetr import *
from .hflayoutlm import *
from .model import *
from .pdftext import *
from .tessocr import *
from .texocr import *  # type: ignore

if tensorpack_available():
    from .tpdetect import *



# -*- coding: utf-8 -*-
# File: base.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module for the base class of datasets.
"""

import os
import pprint
from abc import ABC, abstractmethod
from collections import defaultdict
from typing import Dict, List, Mapping, Optional, Sequence, Tuple, Type, Union

import numpy as np

from ..dataflow import CacheData, ConcatData, CustomDataFromList, DataFlow
from ..datapoint import Image
from ..utils.detection_types import Pathlike
from ..utils.logger import logger
from ..utils.settings import ObjectTypes, TypeOrStr, get_type
from .dataflow_builder import DataFlowBaseBuilder
from .info import DatasetCategories, DatasetInfo, get_merged_categories


class DatasetBase(ABC):
    """
    Base class for a dataset. Requires to implementing `_categories` `_info` and `_builder` by
    yourself. These methods must return a DatasetCategories, a DatasetInfo and a DataFlow_Builder instance, which
    together give a complete description of the dataset. Compare some specific dataset cards in the :mod:`instance` .
    """

    def __init__(self) -> None:
        assert self._info() is not None, "Dataset requires at least a name defined in DatasetInfo"
        self._dataset_info = self._info()
        self._dataflow_builder = self._builder()
        self._dataflow_builder.categories = self._categories()
        self._dataflow_builder.splits = self._dataset_info.splits

        if not self.dataset_available() and self.is_built_in():
            print(
                f"Dataset {self._dataset_info.name} not locally found. Please download at {self._dataset_info.url}"
                f" and place under {self._dataflow_builder.get_workdir()}"
            )

    @property
    def dataset_info(self) -> DatasetInfo:
        """
        dataset_info
        """
        return self._dataset_info

    @property
    def dataflow(self) -> DataFlowBaseBuilder:
        """
        dataflow
        """
        return self._dataflow_builder

    @abstractmethod
    def _categories(self) -> DatasetCategories:
        """
        Construct the DatasetCategory object.
        """

        raise NotImplementedError

    @classmethod
    @abstractmethod
    def _info(cls) -> DatasetInfo:
        """
        Construct the DatasetInfo object.
        """

        raise NotImplementedError

    @abstractmethod
    def _builder(self) -> DataFlowBaseBuilder:
        """
        Construct the DataFlowBaseBuilder object. It needs to be implemented in the derived class.
        """

        raise NotImplementedError

    def dataset_available(self) -> bool:
        """
        Datasets must be downloaded and maybe unzipped manually. Checks, if the folder exists, where the dataset is
        expected.
        """
        if os.path.isdir(self._dataflow_builder.get_workdir()):
            return True
        return False

    @staticmethod
    def is_built_in() -> bool:
        """
        Returns flag to indicate if dataset is custom or built int.
        """
        return False


class _BuiltInDataset(DatasetBase, ABC):
    """
    Dataclass for built-in dataset. Do not use this it
    """

    _name: Optional[str] = None

    @staticmethod
    def is_built_in() -> bool:
        """
        Overwritten from base class
        """
        return True


class SplitDataFlow(DataFlowBaseBuilder):
    """
    Dataflow builder for splitting datasets
    """

    def __init__(self, train: List[Image], val: List[Image], test: Optional[List[Image]]):
        """
        :param train: Cached train split
        :param val: Cached val split
        :param test: Cached test split
        """
        super().__init__(location="")
        self.split_cache: Dict[str, List[Image]]
        if test is None:
            self.split_cache = {"train": train, "val": val}
        else:
            self.split_cache = {"train": train, "val": val, "test": test}

    def build(self, **kwargs: Union[str, int]) -> DataFlow:
        """
        Dataflow builder for merged split datasets
        :param kwargs: Only split and max_datapoints arguments will be considered.
        :return: Dataflow
        """

        split = kwargs.get("split", "train")
        if not isinstance(split, str):
            raise ValueError("'split' must be a string")
        max_datapoints = kwargs.get("max_datapoints")
        if isinstance(max_datapoints, str):
            max_datapoints = int(max_datapoints)

        return CustomDataFromList(self.split_cache[split], max_datapoints=max_datapoints)


class MergeDataset(DatasetBase):
    """
    A class for merging dataset ready to feed a training or an evaluation script. The dataflow builder will generate
    samples from all datasets and will exhaust if every dataflow of the merged datasets are exhausted as well. To
    guarantee flexibility it is possible to pass customized dataflows explicitly to maybe reduce the dataflow size from
    one dataset or to use different splits from different datasets.

    When yielding datapoint from :build(), note that one dataset will pass all its samples successively which
    might reduce randomness for training, especially when using datasets from the same domain. Buffering all datasets
    (without loading heavy components like images) is therefore possible and the merged dataset can be shuffled.

    When the datasets are buffered are split functionality can divide the buffered samples into an train, val and test
    set.

    While the selection of categories is given by the union of all categories of all datasets, sub categories need to
    be handled with care: Only sub categories for one specific category are available provided that every dataset has
    this sub category available for this specific category. The range of sub category values again is defined as the
    range of all values from all datasets.

    **Example:**

            dataset_1 = get_dataset("dataset_1")
            dataset_2 = get_dataset("dataset_2")

            union_dataset = MergeDataset(dataset_1,dataset_2)
            union_dataset.buffer_datasets(split="train")     # will cache the train split of dataset_1 and dataset_2
            merge.split_datasets(ratio=0.1, add_test=False)  # will create a new split of the union.


    **Example:**

            dataset_1 = get_dataset("dataset_1")
            dataset_2 = get_dataset("dataset_2")

            df_1 = dataset_1.dataflow.build(max_datapoints=20)  # handle separate dataflow configs ...
            df_2 = dataset_1.dataflow.build(max_datapoints=30)

            union_dataset = MergeDataset(dataset_1,dataset_2)
            union_dataset.explicit_dataflows(df_1,df_2)   # ... and pass them explicitly. Filtering is another
                                                          # possibility
    """

    def __init__(self, *datasets: DatasetBase):
        """
        :param datasets: An arbitrary number of datasets
        """
        self.datasets = datasets
        self.dataflows: Optional[Tuple[DataFlow, ...]] = None
        self.datapoint_list: Optional[List[Image]] = None
        super().__init__()
        self._dataset_info.type = datasets[0].dataset_info.type
        self._dataset_info.name = "merge_" + "_".join([dataset.dataset_info.name for dataset in self.datasets])

    def _categories(self) -> DatasetCategories:
        return get_merged_categories(
            *(dataset.dataflow.categories for dataset in self.datasets if dataset.dataflow.categories is not None)
        )

    @classmethod
    def _info(cls) -> DatasetInfo:
        return DatasetInfo(name="merge")

    def _builder(self) -> DataFlowBaseBuilder:
        class MergeDataFlow(DataFlowBaseBuilder):
            """
            Dataflow builder for merged datasets
            """

            def __init__(self, *dataflow_builders: DataFlowBaseBuilder):
                super().__init__("")
                self.dataflow_builders = dataflow_builders
                self.dataflows: Optional[Tuple[DataFlow, ...]] = None

            def build(self, **kwargs: Union[str, int]) -> DataFlow:
                """
                Building the dataflow of merged datasets. No argument will affect the stream if the dataflows have
                been explicitly passed. Otherwise, all kwargs will be passed to all dataflows. Note that each dataflow
                will iterate until it is exhausted. To guarantee randomness across different datasets cache all
                datapoints and shuffle them afterwards (e.g. use :buffer_dataset() ).

                :param kwargs: arguments for :build()
                :return: Dataflow
                """
                df_list = []
                if self.dataflows is not None:
                    logger.info("Will used dataflow from previously explicitly passed configuration")
                    return ConcatData(list(self.dataflows))

                logger.info("Will use the same build setting for all dataflows")
                for dataflow_builder in self.dataflow_builders:
                    df_list.append(dataflow_builder.build(**kwargs))
                df = ConcatData(df_list)
                return df

        builder = MergeDataFlow(*(dataset.dataflow for dataset in self.datasets))
        if self.dataflows is not None:
            builder.dataflows = self.dataflows
        return builder

    def explicit_dataflows(self, *dataflows: DataFlow) -> None:
        """
        Pass explicit dataflows for each dataset. Using several dataflow configurations for one dataset is possible as
        well. However, the number of dataflow must exceed the number of merged datasets.

        :param dataflows: An arbitrary number of dataflows
        """
        self.dataflows = dataflows
        if len(self.datasets) > len(self.dataflows):
            raise ValueError(
                f"len(self.datasets) = {len(self.datasets)} must be" f" <= len(self.dataflows) = {len(self.dataflows)}"
            )
        self._dataflow_builder = self._builder()
        self._dataflow_builder.categories = self._categories()

    def buffer_datasets(self, **kwargs: Union[str, int]) -> None:
        """
        Buffer datasets with given configs. If dataflows are passed explicitly it will cache their streamed output.

        :param kwargs: arguments for :build()
        :return: Dataflow
        """
        df = self.dataflow.build(**kwargs)
        self.datapoint_list = CacheData(df, shuffle=True).get_cache()

    def split_datasets(self, ratio: float = 0.1, add_test: bool = True) -> None:
        """
        Split cached datasets into train/val(/test).

        :param ratio: 1-ratio will be assigned to the train split. The remaining bit will be assigned to val and test
                      split.
        :param add_test: Add a test split
        """
        assert self.datapoint_list is not None, "Datasets need to be buffered before splitting"
        number_datapoints = len(self.datapoint_list)
        indices = np.random.binomial(1, ratio, number_datapoints)
        train_dataset = [self.datapoint_list[i] for i in range(number_datapoints) if indices[i] == 0]
        val_dataset = [self.datapoint_list[i] for i in range(number_datapoints) if indices[i] == 1]
        test_dataset = None

        if add_test:
            test_dataset = [dp for id, dp in enumerate(val_dataset) if id % 2]
            val_dataset = [dp for id, dp in enumerate(val_dataset) if not id % 2]

        logger.info("___________________ Number of datapoints per split ___________________")
        logger.info(
            pprint.pformat(
                {
                    "train": len(train_dataset),
                    "val": len(val_dataset),
                    "test": len(test_dataset) if test_dataset is not None else 0,
                },
                width=100,
                compact=True,
            )
        )

        self._dataflow_builder = SplitDataFlow(train_dataset, val_dataset, test_dataset)
        self._dataflow_builder.categories = self._categories()

    def get_ids_by_split(self) -> Dict[str, List[str]]:
        """
        To reproduce a dataset split at a later stage, get a summary of the by having a dict of list with split and
        the image ids contained in the split.

        :return: E.g. `{"train": ['ab','ac'],"val":['bc','bd']}`
        """
        if isinstance(self._dataflow_builder, SplitDataFlow):
            return {
                key: [img.image_id for img in self._dataflow_builder.split_cache.get(key, [])]
                for key in ("train", "val", "test")
            }
        return {"train": [], "val": [], "test": []}

    def create_split_by_id(
        self, split_dict: Mapping[str, Sequence[str]], **dataflow_build_kwargs: Union[str, int]
    ) -> None:
        """
        Reproducing a dataset split from a dataset or a dataflow by a dict of list of image ids.

            merge = dd.MergeDataset(doclaynet)
            merge.explicit_dataflows(df_doc)
            merge.buffer_datasets()
            merge.split_datasets(ratio=0.1)
            out = merge.get_ids_by_split()   # Save out somewhere

            merge_2 = dd.MergeDataset(doclaynet)
            df_doc_2 = doclaynet.dataflow.build(split="train", max_datapoints=4000)
            merge_2.explicit_dataflows(df_doc_2)
            merge_2.create_split_by_id(out)   # merge_2 now has the same split as merge

        :param split_dict: e.g. `{"train":['ab','ac',...],"val":['bc'],"test":[]}`
        """

        if set(split_dict.keys()) != {"train", "val", "test"}:
            raise KeyError("split_dict must contain keys for 'train', 'val' and 'test'")
        ann_id_to_split = {ann_id: "train" for ann_id in split_dict["train"]}
        ann_id_to_split.update({ann_id: "val" for ann_id in split_dict["val"]})
        ann_id_to_split.update({ann_id: "test" for ann_id in split_dict["test"]})
        self.buffer_datasets(**dataflow_build_kwargs)
        split_defaultdict = defaultdict(list)
        for image in self.datapoint_list:  # type: ignore
            split_defaultdict[ann_id_to_split[image.image_id]].append(image)
        train_dataset = split_defaultdict["train"]
        val_dataset = split_defaultdict["val"]
        test_dataset = split_defaultdict["test"]
        self._dataflow_builder = SplitDataFlow(train_dataset, val_dataset, test_dataset)
        self._dataflow_builder.categories = self._categories()


class CustomDataset(DatasetBase):
    """
    A simple dataset interface that implements the boilerplate code and reduces complexity by merely leaving
    the user to write a `DataFlowBaseBuilder` (mapping the annotation format into deepdoctection data model is
    something that has to be left to the user for obvious reasons). Check the tutorial on how to approach the mapping
    problem.
    """

    def __init__(
        self,
        name: str,
        dataset_type: TypeOrStr,
        location: Pathlike,
        init_categories: Sequence[ObjectTypes],
        dataflow_builder: Type[DataFlowBaseBuilder],
        init_sub_categories: Optional[Mapping[ObjectTypes, Mapping[ObjectTypes, Sequence[ObjectTypes]]]] = None,
        annotation_files: Optional[Mapping[str, Union[str, Sequence[str]]]] = None,
    ):
        """
        :param name: Name of the dataset. It will not be used in the code, however it might be helpful, if several
                     custom datasets are in use.
        :param dataset_type: Datasets need to be characterized by one of the `enum` members `DatasetType` that describe
                     the machine learning task the dataset is built for. You can get all registered types with

                            types = dd.object_types_registry.get("DatasetType")
                            print({t for t in types})

        :param location: Datasets should be stored a sub folder of name `location` in the local cache
                         `get_dataset_dir_path()`. There are good reasons to use `name`.
        :param init_categories: A list of all available categories in this dataset. You must use a list as the order
                                of the categories must always be preserved: they determine the category id that in turn
                                will be used for model training.
        :param dataflow_builder: A subclass of `DataFlowBaseBuilder`. Do not instantiate the class by yourself.
        :param init_sub_categories: A dict mapping main categories to sub categories, if there are any available.
                                    Suppose an object `LayoutType.cell` has two additional information in the annotation
                                    file: `CellType.header, CellType.body`. You can then write:

                                        {LayoutType.cell: {CellType.header: [CellType.header, CellType.body]}

                                    This setting assumes that later in the mapping the `ImageAnnotation` with
                                    `category_name=LayoutType.cell` will have a sub category of key `CellType.header`
                                    and one of the two values `CellType.header, CellType.body`
        :param annotation_files: A mapping to one or more annotation files, e.g.

                                       annotation_file = {"train": "train_file.json", "test": "test_file.json"}
        """

        self.name = name
        self.type = get_type(dataset_type)
        self.location = location
        self.init_categories = init_categories
        if init_sub_categories is None:
            self.init_sub_categories: Mapping[ObjectTypes, Mapping[ObjectTypes, Sequence[ObjectTypes]]] = {}
        else:
            self.init_sub_categories = init_sub_categories
        self.annotation_files = annotation_files
        self.dataflow_builder = dataflow_builder(self.location, self.annotation_files)
        super().__init__()

    def _info(self) -> DatasetInfo:  # type: ignore  # pylint: disable=W0221
        return DatasetInfo(name=self.name, type=self.type, description="", license="", url="", splits={})

    def _categories(self) -> DatasetCategories:
        return DatasetCategories(init_categories=self.init_categories, init_sub_categories=self.init_sub_categories)

    def _builder(self) -> DataFlowBaseBuilder:
        return self.dataflow_builder



# -*- coding: utf-8 -*-
# File: hf_layoutlm_train.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module for training Huggingface implementation of LayoutLm
"""

import copy
import json
import pprint
from typing import Any, Dict, List, Mapping, Optional, Sequence, Tuple, Type, Union

from torch.nn import Module
from torch.utils.data import Dataset
from transformers import (
    IntervalStrategy,
    LayoutLMForSequenceClassification,
    LayoutLMForTokenClassification,
    LayoutLMTokenizerFast,
    LayoutLMv2Config,
    LayoutLMv2ForSequenceClassification,
    LayoutLMv2ForTokenClassification,
    LayoutLMv3Config,
    LayoutLMv3ForSequenceClassification,
    LayoutLMv3ForTokenClassification,
    PretrainedConfig,
    PreTrainedModel,
    RobertaTokenizerFast,
    XLMRobertaTokenizerFast,
)
from transformers.trainer import Trainer, TrainingArguments

from ..datasets.adapter import DatasetAdapter
from ..datasets.base import DatasetBase
from ..datasets.registry import get_dataset
from ..eval.accmetric import ClassificationMetric
from ..eval.eval import Evaluator
from ..extern.hflayoutlm import (
    HFLayoutLmSequenceClassifier,
    HFLayoutLmTokenClassifier,
    HFLayoutLmv2SequenceClassifier,
    HFLayoutLmv2TokenClassifier,
    HFLayoutLmv3SequenceClassifier,
    HFLayoutLmv3TokenClassifier,
)
from ..mapper.laylmstruct import LayoutLMDataCollator, image_to_raw_layoutlm_features
from ..pipe.base import LanguageModelPipelineComponent
from ..pipe.lm import get_tokenizer_from_architecture
from ..pipe.registry import pipeline_component_registry
from ..utils.logger import logger
from ..utils.settings import DatasetType, LayoutType, ObjectTypes, WordType
from ..utils.utils import string_to_dict

_ARCHITECTURES_TO_MODEL_CLASS = {
    "LayoutLMForTokenClassification": (LayoutLMForTokenClassification, HFLayoutLmTokenClassifier, PretrainedConfig),
    "LayoutLMForSequenceClassification": (
        LayoutLMForSequenceClassification,
        HFLayoutLmSequenceClassifier,
        PretrainedConfig,
    ),
    "LayoutLMv2ForTokenClassification": (
        LayoutLMv2ForTokenClassification,
        HFLayoutLmv2TokenClassifier,
        LayoutLMv2Config,
    ),
    "LayoutLMv2ForSequenceClassification": (
        LayoutLMv2ForSequenceClassification,
        HFLayoutLmv2SequenceClassifier,
        LayoutLMv2Config,
    ),
}


_MODEL_TYPE_AND_TASK_TO_MODEL_CLASS: Mapping[Tuple[str, ObjectTypes], Any] = {
    ("layoutlm", DatasetType.sequence_classification): (
        LayoutLMForSequenceClassification,
        HFLayoutLmSequenceClassifier,
        PretrainedConfig,
    ),
    ("layoutlm", DatasetType.token_classification): (
        LayoutLMForTokenClassification,
        HFLayoutLmTokenClassifier,
        PretrainedConfig,
    ),
    ("layoutlmv2", DatasetType.sequence_classification): (
        LayoutLMv2ForSequenceClassification,
        HFLayoutLmv2SequenceClassifier,
        LayoutLMv2Config,
    ),
    ("layoutlmv2", DatasetType.token_classification): (
        LayoutLMv2ForTokenClassification,
        HFLayoutLmv2TokenClassifier,
        LayoutLMv2Config,
    ),
    ("layoutlmv3", DatasetType.sequence_classification): (
        LayoutLMv3ForSequenceClassification,
        HFLayoutLmv3SequenceClassifier,
        LayoutLMv3Config,
    ),
    ("layoutlmv3", DatasetType.token_classification): (
        LayoutLMv3ForTokenClassification,
        HFLayoutLmv3TokenClassifier,
        LayoutLMv3Config,
    ),
}
_MODEL_TYPE_TO_TOKENIZER = {
    ("layoutlm", False): LayoutLMTokenizerFast.from_pretrained("microsoft/layoutlm-base-uncased"),
    ("layoutlmv2", False): LayoutLMTokenizerFast.from_pretrained("microsoft/layoutlm-base-uncased"),
    ("layoutlmv2", True): XLMRobertaTokenizerFast.from_pretrained("xlm-roberta-base", add_prefix_space=True),
    ("layoutlmv3", False): RobertaTokenizerFast.from_pretrained("roberta-base", add_prefix_space=True),
}


class LayoutLMTrainer(Trainer):
    """
    Huggingface Trainer for training Transformer models with a custom evaluate method in order
    to use dd Evaluator. Train setting is not defined in the trainer itself but in config setting as
    defined in `TrainingArguments`. Please check the Transformer documentation

    <https://huggingface.co/docs/transformers/main_classes/trainer>

    for custom training setting.
    """

    def __init__(
        self,
        model: Union[PreTrainedModel, Module],
        args: TrainingArguments,
        data_collator: LayoutLMDataCollator,
        train_dataset: Dataset[Any],
    ):
        self.evaluator: Optional[Evaluator] = None
        self.build_eval_kwargs: Optional[Dict[str, Any]] = None
        super().__init__(model, args, data_collator, train_dataset)

    def setup_evaluator(
        self,
        dataset_val: DatasetBase,
        pipeline_component: LanguageModelPipelineComponent,
        metric: Union[Type[ClassificationMetric], ClassificationMetric],
        **build_eval_kwargs: Union[str, int],
    ) -> None:
        """
        Setup of evaluator before starting training. During training, predictors will be replaced by current
        checkpoints.

        :param dataset_val: dataset on which to run evaluation
        :param pipeline_component: pipeline component to plug into the evaluator
        :param metric: A metric class
        :param build_eval_kwargs:
        """

        self.evaluator = Evaluator(dataset_val, pipeline_component, metric, num_threads=1)
        assert self.evaluator.pipe_component
        for comp in self.evaluator.pipe_component.pipe_components:
            comp.language_model.model = None  # type: ignore
        self.build_eval_kwargs = build_eval_kwargs

    def evaluate(
        self,
        eval_dataset: Optional[Dataset[Any]] = None,
        ignore_keys: Optional[List[str]] = None,
        metric_key_prefix: str = "eval",
    ) -> Dict[str, float]:
        """
        Overwritten method from `Trainer`. Arguments will not be used.
        """
        assert self.evaluator is not None
        assert self.evaluator.pipe_component is not None

        # memory metrics - must set up as early as possible
        self._memory_tracker.start()
        for comp in self.evaluator.pipe_component.pipe_components:
            comp.language_model.model = copy.deepcopy(self.model).eval()  # type: ignore
        if isinstance(self.build_eval_kwargs, dict):
            scores = self.evaluator.run(True, **self.build_eval_kwargs)
        else:
            scores = self.evaluator.run(True)

        self.log(scores)

        return scores


def _get_model_class_and_tokenizer(
    path_config_json: str, dataset_type: ObjectTypes, use_xlm_tokenizer: bool
) -> Tuple[Any, Any, Any, Any]:
    with open(path_config_json, "r", encoding="UTF-8") as file:
        config_json = json.load(file)

    model_type = config_json.get("model_type")

    if architectures := config_json.get("architectures"):
        model_cls, model_wrapper_cls, config_cls = _ARCHITECTURES_TO_MODEL_CLASS[architectures[0]]
        tokenizer_fast = get_tokenizer_from_architecture(architectures[0], use_xlm_tokenizer)
    elif model_type:
        model_cls, model_wrapper_cls, config_cls = _MODEL_TYPE_AND_TASK_TO_MODEL_CLASS[(model_type, dataset_type)]
        tokenizer_fast = _MODEL_TYPE_TO_TOKENIZER[(model_type, use_xlm_tokenizer)]
    else:
        raise KeyError("model_type and architectures not available in configs")

    if not model_cls:
        raise ValueError("model not eligible to run with this framework")

    return config_cls, model_cls, model_wrapper_cls, tokenizer_fast


def train_hf_layoutlm(
    path_config_json: str,
    dataset_train: Union[str, DatasetBase],
    path_weights: str,
    config_overwrite: Optional[List[str]] = None,
    log_dir: str = "train_log/layoutlm",
    build_train_config: Optional[Sequence[str]] = None,
    dataset_val: Optional[DatasetBase] = None,
    build_val_config: Optional[Sequence[str]] = None,
    metric: Optional[Union[Type[ClassificationMetric], ClassificationMetric]] = None,
    pipeline_component_name: Optional[str] = None,
    use_xlm_tokenizer: bool = False,
    use_token_tag: bool = True,
    segment_positions: Optional[Union[LayoutType, Sequence[LayoutType]]] = None,
) -> None:
    """
    Script for fine-tuning LayoutLM models either for sequence classification (e.g. classifying documents) or token
    classification using HF Trainer and custom evaluation. It currently supports LayoutLM, LayoutLMv2, LayoutLMv3 and
    LayoutXLM. Training similar but different models like LILT <https://arxiv.org/abs/2202.13669> can be done by
    changing a few lines of code regarding the selection of the tokenizer.

    The theoretical foundation can be taken from

    <https://arxiv.org/abs/1912.13318>

    This is not the pre-training script.

    In order to remain within the framework of this library, the base and uncased LayoutLM model must be downloaded
    from the HF-hub in a first step for fine-tuning.  Models are available for this, which are registered in the
    ModelCatalog. It is possible to choose one of the following options:

        "microsoft/layoutlm-base-uncased/pytorch_model.bin"
        "microsoft/layoutlmv2-base-uncased/pytorch_model.bin"
        "microsoft/layoutxlm-base/pytorch_model.bin"
        "microsoft/layoutlmv3-base/pytorch_model.bin"

     and

         "microsoft/layoutlm-large-uncased/pytorch_model.bin"

    (You can also choose the large versions of LayoutLMv2 and LayoutXLM but you need to organize the download yourself.)

        ModelDownloadManager.maybe_download_weights_and_configs("microsoft/layoutlm-base-uncased/pytorch_model.bin")

    The corresponding cased models are currently not available, but this is only to keep the model selection small.

    If the config file and weights have been downloaded, the model can be trained for the desired task.

    How does the model selection work?

    The base model is selected by the transferred config file and the weights. Depending on the dataset type
    ("SEQUENCE_CLASSIFICATION" or "TOKEN_CLASSIFICATION"), the complete model is then put together by placing a suitable
    top layer on the base model.

    :param path_config_json: Absolute path to HF config file, e.g.
                             ModelCatalog.get_full_path_configs("microsoft/layoutlm-base-uncased/pytorch_model.bin")
    :param dataset_train: Dataset to use for training. Only datasets of type "SEQUENCE_CLASSIFICATION" or
                          "TOKEN_CLASSIFICATION" are supported.
    :param path_weights: path to a checkpoint for further fine-tuning
    :param config_overwrite: Pass a list of arguments if some configs from `TrainingArguments` should be replaced. Check
                             https://huggingface.co/docs/transformers/main_classes/trainer#transformers.TrainingArguments
                             for the full training default setting.
    :param log_dir: Path to log dir. Will default to `train_log/layoutlm`
    :param build_train_config: dataflow build setting. Again, use list convention setting, e.g. ['max_datapoints=1000']
    :param dataset_val: Dataset to use for validation. Dataset type must be the same as type of `dataset_train`
    :param build_val_config: same as `build_train_config` but for validation
    :param metric: A metric to choose for validation.
    :param pipeline_component_name: A pipeline component name to use for validation (e.g. LMSequenceClassifierService or
                                    LMTokenClassifierService.
    :param use_xlm_tokenizer: This is only necessary if you pass weights of layoutxlm. The config cannot distinguish
                              between Layoutlmv2 and Layoutxlm, so you need to pass this info explicitly.
    :param use_token_tag: Will only be used for dataset_type="token_classification". If use_token_tag=True, will use
                          labels from sub category `WordType.token_tag` (with `B,I,O` suffix), otherwise
                          `WordType.token_class`.
    :param segment_positions: Using bounding boxes of segment instead of words improves model accuracy significantly.
                              Choose a single or a sequence of layout segments to use their bounding boxes. Note, that
                              the layout segments need to have a child-relationship with words. If a word does not
                              appear as child, it will use the word bounding box.
    """

    build_train_dict: Dict[str, str] = {}
    if build_train_config is not None:
        build_train_dict = string_to_dict(",".join(build_train_config))
    if "split" not in build_train_dict:
        build_train_dict["split"] = "train"

    build_val_dict: Dict[str, str] = {}
    if build_val_config is not None:
        build_val_dict = string_to_dict(",".join(build_val_config))
    if "split" not in build_val_dict:
        build_val_dict["split"] = "val"

    if config_overwrite is None:
        config_overwrite = []

    if isinstance(dataset_train, str):
        dataset_train = get_dataset(dataset_train)

    # We wrap our dataset into a torch dataset
    dataset_type = dataset_train.dataset_info.type
    if dataset_type == DatasetType.sequence_classification:
        categories_dict_name_as_key = dataset_train.dataflow.categories.get_categories(as_dict=True, name_as_key=True)
    elif dataset_type == DatasetType.token_classification:
        if use_token_tag:
            categories_dict_name_as_key = dataset_train.dataflow.categories.get_sub_categories(
                categories=LayoutType.word,
                sub_categories={LayoutType.word: [WordType.token_tag]},
                keys=False,
                values_as_dict=True,
                name_as_key=True,
            )[LayoutType.word][WordType.token_tag]
        else:
            categories_dict_name_as_key = dataset_train.dataflow.categories.get_sub_categories(
                categories=LayoutType.word,
                sub_categories={LayoutType.word: [WordType.token_class]},
                keys=False,
                values_as_dict=True,
                name_as_key=True,
            )[LayoutType.word][WordType.token_class]
    else:
        raise ValueError("Dataset type not supported for training")

    config_cls, model_cls, model_wrapper_cls, tokenizer_fast = _get_model_class_and_tokenizer(
        path_config_json, dataset_type, use_xlm_tokenizer
    )
    image_to_raw_layoutlm_kwargs = {"dataset_type": dataset_type, "use_token_tag": use_token_tag}
    if segment_positions:
        image_to_raw_layoutlm_kwargs["segment_positions"] = segment_positions
    image_to_raw_layoutlm_kwargs.update(model_wrapper_cls.default_kwargs_for_input_mapping())
    dataset = DatasetAdapter(
        dataset_train,
        True,
        image_to_raw_layoutlm_features(**image_to_raw_layoutlm_kwargs),
        use_token_tag,
        **build_train_dict,
    )

    number_samples = len(dataset)
    # A setup of necessary configuration. Everything else will be equal to the default setting of the transformer
    # library.
    # Need to set remove_unused_columns to False, as the DataCollator for column removal will remove some raw features
    # that are necessary for the tokenizer.
    conf_dict = {
        "output_dir": log_dir,
        "remove_unused_columns": False,
        "per_device_train_batch_size": 8,
        "max_steps": number_samples,
        "evaluation_strategy": "steps"
        if (dataset_val is not None and metric is not None and pipeline_component_name is not None)
        else "no",
        "eval_steps": 100,
    }

    # We allow to overwrite the default setting by the user.
    for conf in config_overwrite:
        key, val = conf.split("=", maxsplit=1)
        try:
            val = int(val)  # type: ignore
        except ValueError:
            try:
                val = float(val)  # type: ignore
            except ValueError:
                pass
        conf_dict[key] = val

    # Will inform about dataloader warnings if max_steps exceeds length of dataset
    if conf_dict["max_steps"] > number_samples:  # type: ignore
        logger.warning(
            "After %s dataloader will log warning at every iteration about unexpected samples", number_samples
        )

    arguments = TrainingArguments(**conf_dict)
    logger.info("Config: \n %s", str(arguments.to_dict()), arguments.to_dict())

    id2label = {int(k) - 1: v for v, k in categories_dict_name_as_key.items()}

    logger.info("Will setup a head with the following classes\n %s", pprint.pformat(id2label, width=100, compact=True))

    config = config_cls.from_pretrained(pretrained_model_name_or_path=path_config_json, id2label=id2label)
    model = model_cls.from_pretrained(pretrained_model_name_or_path=path_weights, config=config)
    data_collator = LayoutLMDataCollator(tokenizer_fast, return_tensors="pt")
    trainer = LayoutLMTrainer(model, arguments, data_collator, dataset)

    if arguments.evaluation_strategy in (IntervalStrategy.STEPS,):
        assert metric is not None  # silence mypy
        if dataset_type == DatasetType.sequence_classification:
            categories = dataset_val.dataflow.categories.get_categories(filtered=True)  # type: ignore
        else:
            if use_token_tag:
                categories = dataset_val.dataflow.categories.get_sub_categories(  # type: ignore
                    categories=LayoutType.word, sub_categories={LayoutType.word: [WordType.token_tag]}, keys=False
                )[LayoutType.word][WordType.token_tag]
                metric.set_categories(category_names=LayoutType.word, sub_category_names={"word": ["token_tag"]})
            else:
                categories = dataset_val.dataflow.categories.get_sub_categories(  # type: ignore
                    categories=LayoutType.word, sub_categories={LayoutType.word: [WordType.token_class]}, keys=False
                )[LayoutType.word][WordType.token_class]
                metric.set_categories(category_names=LayoutType.word, sub_category_names={"word": ["token_class"]})
        dd_model = model_wrapper_cls(
            path_config_json=path_config_json,
            path_weights=path_weights,
            categories=categories,
            device="cuda",
        )
        pipeline_component_cls = pipeline_component_registry.get(pipeline_component_name)
        if dataset_type == DatasetType.sequence_classification:
            pipeline_component = pipeline_component_cls(tokenizer_fast, dd_model)
        else:
            pipeline_component = pipeline_component_cls(tokenizer_fast, dd_model, use_other_as_default_category=True)
        assert isinstance(pipeline_component, LanguageModelPipelineComponent)

        trainer.setup_evaluator(dataset_val, pipeline_component, metric, **build_val_dict)  # type: ignore

    trainer.train()



# -*- coding: utf-8 -*-
# File: test_common.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Testing the module dataflow.common
"""

from typing import List, no_type_check

import numpy as np
from numpy.testing import assert_array_equal
from pytest import mark

from deepdoctection.dataflow import (
    ConcatData,
    CustomDataFromList,
    DataFlow,
    DataFromList,
    FakeData,
    FlattenData,
    JoinData,
    MapData,
    MapDataComponent,
    RepeatedData,
)

from ..test_utils import collect_datapoint_from_dataflow


@mark.basic
def test_flatten_data(dataset_three_dim: List[List[float]], dataset_flatten: List[List[List[float]]]) -> None:
    """
    Test the flattening of a dataflow
    """

    # Arrange
    df: DataFlow
    df = CustomDataFromList(dataset_three_dim)

    # Act
    df = FlattenData(df)
    output = collect_datapoint_from_dataflow(df=df)

    # Assert
    assert output == dataset_flatten


@mark.basic
def test_map_data() -> None:
    """Test MapData"""

    # Arrange
    @no_type_check
    def map_to_one(dp):
        return np.ones(dp[0].shape)

    df: DataFlow
    df = FakeData(shapes=[[4, 7, 3]], domain=(0, 1), size=1)

    # Act
    df = MapData(df, map_to_one)
    output = collect_datapoint_from_dataflow(df=df)

    # Assert
    assert_array_equal(output[0], np.ones((4, 7, 3)))


@mark.basic
def test_map_data_component() -> None:
    """Test MapDataComponent"""

    # Arrange
    dataflow_list = [{"foo": "1", "bak": "a"}, {"foo": "3", "bak": "c"}]
    df: DataFlow
    df = DataFromList(dataflow_list, shuffle=False)

    # Act
    df = MapDataComponent(df, lambda dp: "4", "foo")
    output = collect_datapoint_from_dataflow(df=df)

    # Assert
    assert output[0]["foo"] == "4"
    assert output[0]["bak"] == "a"
    assert output[1]["foo"] == "4"
    assert output[1]["bak"] == "c"


@mark.basic
def test_repeated_data() -> None:
    """Test RepeatedData"""

    # Arrange
    dataflow_list = [{"foo": "1", "bak": "a"}, {"foo": "3", "bak": "c"}]
    df: DataFlow
    df = DataFromList(dataflow_list, shuffle=False)

    # Act
    df = RepeatedData(df, num=2)
    output = collect_datapoint_from_dataflow(df=df)

    # Arrange
    assert len(output) == 4


@mark.basic
def test_repeated_data_yields_infinitely_many_datapoints() -> None:
    """Test RepeatedData produces infinitely many datapoints"""
    # Arrange
    dataflow_list = [{"foo": "1", "bak": "a"}, {"foo": "3", "bak": "c"}]
    df: DataFlow
    df = DataFromList(dataflow_list, shuffle=False)

    # Act
    df = RepeatedData(df, -1)
    output = collect_datapoint_from_dataflow(df=df, max_datapoints=100)

    # Assert
    assert len(output) == 100


@mark.basic
def test_concat_data() -> None:
    """Test ConcatData"""

    # Arrange
    dataflow_list_1 = [{"foo": "1", "bak": "a"}, {"foo": "3", "bak": "c"}]
    dataflow_list_2 = [{"foo": "2", "bak": "a"}, {"foo": "4", "bak": "c"}]
    df_1 = DataFromList(dataflow_list_1, shuffle=False)
    df_2 = DataFromList(dataflow_list_2, shuffle=False)

    # Act
    df: DataFlow
    df = ConcatData([df_1, df_2])
    output = collect_datapoint_from_dataflow(df=df)

    # Assert
    assert len(output) == 4


@mark.basic
def test_join_data() -> None:
    """Test JoinData"""

    # Arrange
    dataflow_list_1 = [{"foo": "1", "bak": "a"}, {"foo": "3", "bak": "c"}]
    dataflow_list_2 = [{"baz": "2", "bal": "a"}, {"baz": "4", "bal": "c"}]
    df_1 = DataFromList(dataflow_list_1, shuffle=False)
    df_2 = DataFromList(dataflow_list_2, shuffle=False)

    # Act
    df: DataFlow
    df = JoinData([df_1, df_2])
    output = collect_datapoint_from_dataflow(df=df)

    # Assert
    assert len(output) == 2
    assert "foo" in output[0] and "bak" in output[0] and "baz" in output[0] and "bal" in output[0]



# -*- coding: utf-8 -*-
# File: rvldcip.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Module for Publaynet dataset. Place the dataset as follows

    rvl-cdip
    ├── images
    │ ├── imagesa
    │ ├── ...
    │ ├── imagesa
    ├── label
    │ ├── test.txt
    │ ├── train.txt
    │ ├── val.txt
"""

import os
from typing import Mapping, Union

from ...dataflow import DataFlow, MapData
from ...dataflow.custom_serialize import SerializerTabsepFiles
from ...datapoint.annotation import CategoryAnnotation, SummaryAnnotation
from ...datapoint.image import Image
from ...mapper.cats import filter_summary
from ...mapper.maputils import curry
from ...utils.fs import load_image_from_file
from ...utils.settings import DatasetType, DocumentType, PageType, TypeOrStr
from ..base import _BuiltInDataset
from ..dataflow_builder import DataFlowBaseBuilder
from ..info import DatasetCategories, DatasetInfo
from ..registry import dataset_registry

_NAME = "rvl-cdip"
_DESCRIPTION = (
    "The RVL-CDIP (Ryerson Vision Lab Complex Document Information Processing) dataset consists of 400, 000 gray- \n"
    "scale images in 16 classes, with 25, 000 images per class . There are 320, 000 training images, 40, 000  \n"
    "validation images, and 40, 000 test images.The images are sized so their largest dimension does not exceed 1000 \n"
    "pixels. This dataset is a subset of the IIT-CDIP Test Collection 1.0 [1], which is publicly available here. The \n"
    "file structure of this dataset is the same as in the IIT collection, so it is possible to refer to that dataset \n"
    "for OCR and additional metadata. The IIT-CDIP dataset is itself a subset of the Legacy Tobacco Document Library"
)
_LICENSE = (
    "RVL-CDIP is a subset of IIT-CDIP, which came from the Legacy Tobacco Document Library, for which \n"
    "license information can be found at https://www.industrydocuments.ucsf.edu/help/copyright/ ."
)

_URL = "https://www.cs.cmu.edu/~aharley/rvl-cdip/"

_SPLITS: Mapping[str, str] = {"train": "train", "val": "val", "test": "test"}
_TYPE = DatasetType.sequence_classification
_LOCATION = "rvl-cdip"

_ANNOTATION_FILES: Mapping[str, str] = {"train": "labels/train.txt", "val": "labels/val.txt", "test": "labels/test.txt"}
_INIT_CATEGORIES = [
    DocumentType.letter,
    DocumentType.form,
    DocumentType.email,
    DocumentType.handwritten,
    DocumentType.advertisement,
    DocumentType.scientific_report,
    DocumentType.scientific_publication,
    DocumentType.specification,
    DocumentType.file_folder,
    DocumentType.news_article,
    DocumentType.budget,
    DocumentType.invoice,
    DocumentType.presentation,
    DocumentType.questionnaire,
    DocumentType.resume,
    DocumentType.memo,
]


@dataset_registry.register("rvl-cdip")
class Rvlcdip(_BuiltInDataset):
    """
    RVLCDIP
    """

    _name = _NAME

    @classmethod
    def _info(cls) -> DatasetInfo:
        return DatasetInfo(name=_NAME, description=_DESCRIPTION, license=_LICENSE, url=_URL, splits=_SPLITS, type=_TYPE)

    def _categories(self) -> DatasetCategories:
        return DatasetCategories(init_categories=_INIT_CATEGORIES)

    def _builder(self) -> "RvlcdipBuilder":
        return RvlcdipBuilder(location=_LOCATION, annotation_files=_ANNOTATION_FILES)


class RvlcdipBuilder(DataFlowBaseBuilder):
    """
    Rvlcdip dataflow builder
    """

    def build(self, **kwargs: Union[str, int]) -> DataFlow:
        """
        Returns a dataflow from which you can stream datapoints of images. The following arguments affect the returns
        of the dataflow:

        `split:` Split of the dataset. Can be `train`, `val` or `test`. Default: `val`

        max_datapoints: Will stop iterating after max_datapoints. Default: `None`

        load_image: Will load the image for each datapoint.  Default: `False`

        :return: dataflow
        """

        split = str(kwargs.get("split", "val"))
        max_datapoints = kwargs.get("max_datapoints")
        if max_datapoints is not None:
            max_datapoints = int(max_datapoints)
        load_image = kwargs.get("load_image", False)

        # Load
        df: DataFlow
        path = self.get_workdir() / self.get_annotation_file(split)
        df = SerializerTabsepFiles.load(path, max_datapoints)

        @curry
        def _map_str_to_image(dp: str, load_img: bool) -> Image:
            location, label = dp.split()[0], dp.split()[1]
            label = str(int(label) + 1)
            file_name = os.path.split(location)[1]
            image = Image(location=(self.get_workdir() / "images" / location).as_posix(), file_name=file_name)
            image.image = load_image_from_file(image.location)
            summary = SummaryAnnotation()
            categories_dict = self.categories.get_categories(init=True)
            summary.dump_sub_category(
                PageType.document_type, CategoryAnnotation(category_name=categories_dict[label], category_id=str(label))
            )
            image.summary = summary
            if not load_img:
                image.clear_image()
            return image

        df = MapData(df, _map_str_to_image(load_image))  # pylint: disable=E1120

        if self.categories.is_filtered():
            df = MapData(
                df,
                filter_summary({PageType.document_type: self.categories.get_categories(as_dict=False, filtered=True)}),
            )

            @curry
            def _re_map_cat_ids(dp: Image, filtered_categories_name_as_key: Mapping[TypeOrStr, str]) -> Image:
                if dp.summary:
                    if PageType.document_type in dp.summary.sub_categories:
                        summary_cat = dp.summary.get_sub_category(PageType.document_type)
                        summary_cat.category_id = filtered_categories_name_as_key[summary_cat.category_name]
                return dp

            df = MapData(df, _re_map_cat_ids(self.categories.get_categories(filtered=True, name_as_key=True)))
        return df



# -*- coding: utf-8 -*-
# File: concurrency.py

# Copyright 2022 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Some utility functions for multi threading purposes
"""

import multiprocessing as mp
import platform
import queue
import signal
import sys
import threading
from contextlib import contextmanager
from typing import Any, Generator, Optional, no_type_check

from .detection_types import QueueType
from .logger import log_once


# taken from https://github.com/tensorpack/dataflow/blob/master/dataflow/utils/concurrency.py
class StoppableThread(threading.Thread):
    """
    A thread that has a 'stop' event.
    """

    def __init__(self, evt: Optional[threading.Event] = None) -> None:
        """
        :param evt: if None, will create one.
        """
        super().__init__()
        if evt is None:
            evt = threading.Event()
        self._stop_evt = evt

    def stop(self) -> None:
        """Stop the thread"""
        self._stop_evt.set()

    def stopped(self) -> bool:
        """
        :param bool: whether the thread is stopped or not
        """
        return self._stop_evt.isSet()

    def queue_put_stoppable(self, q: QueueType, obj: Any) -> None:
        """Put obj to queue, but will give up when the thread is stopped"""
        while not self.stopped():
            try:
                q.put(obj, timeout=5)
                break
            except queue.Full:
                pass

    def queue_get_stoppable(self, q: QueueType) -> Any:
        """Take obj from queue, but will give up when the thread is stopped"""
        while not self.stopped():
            try:
                return q.get(timeout=5)
            except queue.Empty:
                pass


@contextmanager
def mask_sigint() -> Generator[Any, None, None]:
    """[Any,None,None
    :return: If called in main thread, returns a context where ``SIGINT`` is ignored, and yield True.
             Otherwise, yield False.
    """
    if threading.current_thread() == threading.main_thread():
        sigint_handler = signal.signal(signal.SIGINT, signal.SIG_IGN)
        yield True
        signal.signal(signal.SIGINT, sigint_handler)
    else:
        yield False


def enable_death_signal(_warn: bool = True) -> None:
    """
    Set the "death signal" of the current process, so that
    the current process will be cleaned with guarantee
    in case the parent dies accidentally.
    """
    if platform.system() != "Linux":
        return
    try:
        import prctl  # type: ignore #pylint: disable=C0415  # pip install python-prctl
    except ImportError:
        if _warn:
            log_once(
                '"import prctl" failed! Install python-prctl so that processes can be cleaned with guarantee.', "warn"
            )
        return
    else:
        assert hasattr(
            prctl, "set_pdeathsig"
        ), "prctl.set_pdeathsig does not exist! Note that you need to install 'python-prctl' instead of 'prctl'."
        # is SIGHUP a good choice?
        prctl.set_pdeathsig(signal.SIGHUP)


# taken from https://github.com/tensorpack/dataflow/blob/master/dataflow/utils/concurrency.py


@no_type_check
def start_proc_mask_signal(proc):
    """
    Start process(es) with SIGINT ignored.

    :param proc: (mp.Process or list)

    The signal mask is only applied when called from main thread.
    """
    if not isinstance(proc, list):
        proc = [proc]

    with mask_sigint():
        for pro in proc:
            if isinstance(pro, mp.Process):
                if sys.version_info < (3, 4) or mp.get_start_method() == "fork":
                    log_once(
                        """
Starting a process with 'fork' method is efficient but not safe and may cause deadlock or crash.
Use 'forkserver' or 'spawn' method instead if you run into such issues.
See https://docs.python.org/3/library/multiprocessing.html#contexts-and-start-methods on how to set them.
""".replace(
                            "\n", ""
                        ),
                        "warn",
                    )  # noqa
            pro.start()



# -*- coding: utf-8 -*-
# File: __init__.py

"""
Init file for utils package
"""
from typing import Optional, Tuple, Union, no_type_check

from .context import *
from .file_utils import *
from .fs import *
from .identifier import *
from .logger import *
from .metacfg import *
from .pdf_utils import *
from .settings import *
from .systools import *
from .tqdm import *
from .transform import *
from .utils import *
from .viz import *

__all__ = []


@no_type_check
def _global_import(
    name, prefix: Optional[Union[str, Tuple[str, ...]]] = None, suffix: Optional[Union[str, Tuple[str, ...]]] = None
):
    prefix_default = prefix is None
    suffix_default = suffix is None
    p = __import__(name, globals(), None, level=1)  # pylint: disable=C0103
    lst = p.__all__ if "__all__" in dir(p) else dir(p)
    for k in lst:
        if not k.startswith("__"):
            if prefix_default and suffix_default:
                globals()[k] = p.__dict__[k]
                __all__.append(k)
            elif not prefix_default:
                if k.startswith(prefix):
                    globals()[k] = p.__dict__[k]
                    __all__.append(k)
            elif not suffix_default:
                if k.endswith(suffix):
                    globals()[k] = p.__dict__[k]
                    __all__.append(k)


_global_import("file_utils", suffix=("_available", "_requirement"))
_global_import("metacfg", prefix=("set_config_by_yaml", "save_config_to_yaml", "config_to_cli_str"))
_global_import("utils", prefix=("delete_keys_from_dict", "split_string", "string_to_dict"))
_global_import(
    "settings", suffix=("Type", "TokenClasses", "BioTag", "TokenClassWithTag", "Relationships", "Languages", "get_type")
)

# pylint: disable=undefined-variable
__all__.extend(context.__all__)  # type: ignore
__all__.extend(fs.__all__)  # type: ignore
__all__.extend(identifier.__all__)  # type: ignore
__all__.extend(["logger", "set_logger_dir", "auto_set_dir", "get_logger_dir"])
__all__.extend(pdf_utils.__all__)  # type: ignore
__all__.extend(systools.__all__)  # type: ignore
__all__.extend(["get_tqdm"])
__all__.extend(transform.__all__)  # type: ignore
__all__.extend(viz.__all__)  # type: ignore
# pylint: enable=undefined-variable



# -*- coding: utf-8 -*-
# File: test_stats.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Testing the module dataflow.stats
"""
from typing import List

import numpy as np
from numpy.testing import assert_array_almost_equal, assert_array_equal
from pytest import mark

from deepdoctection.dataflow import CustomDataFromList, MeanFromDataFlow, StdFromDataFlow


class TestMeanFromDataFlow:
    """
    Testing MeanFromDataFlow along various dimensions
    """

    @staticmethod
    @mark.basic
    def test_mean_along_one_dimension(dataset_three_dim: List[List[float]], mean_axis_zero: List[float]) -> None:
        """
        Test mean along the 0th dimension.
        """

        # Arrange
        df = CustomDataFromList(dataset_three_dim)

        # Act
        mean = MeanFromDataFlow(df, axis=0).start()

        # Assert
        assert_array_equal(mean, np.array(mean_axis_zero))

    @staticmethod
    @mark.basic
    def test_mean_along_all_dimension(dataset_three_dim: List[List[float]], mean_all_axes: List[float]) -> None:
        """
        Test mean along all dimensions.
        """

        # Arrange
        df = CustomDataFromList(dataset_three_dim)

        # Act
        mean = MeanFromDataFlow(df).start()

        # Assert
        assert_array_equal(mean, np.array(mean_all_axes))


class TestStdFromDataFlow:
    """
    Testing StdFromDataFlow along various dimensions
    """

    @staticmethod
    @mark.basic
    def test_std_along_one_dimension(dataset_three_dim: List[List[float]], std_axis_zero: List[float]) -> None:
        """
        Test std along the 0th dimension.
        """

        # Arrange
        df = CustomDataFromList(dataset_three_dim)

        # Act
        std = StdFromDataFlow(df, axis=0).start()

        # Assert
        assert_array_almost_equal(std, np.array(std_axis_zero), decimal=4)

    @staticmethod
    @mark.basic
    def test_std_along_all_dimension(dataset_three_dim: List[List[float]], std_all_axes: List[float]) -> None:
        """
        Test std along all dimensions.
        """

        # Arrange
        df = CustomDataFromList(dataset_three_dim)

        # Act
        std = StdFromDataFlow(df).start()

        # Assert
        assert_array_almost_equal(std, np.array(std_all_axes), decimal=4)



# -*- coding: utf-8 -*-
# File: d2struct.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module for mapping annotations into standard Detectron2 dataset dict. Also providing some tools for W&B mapping and
visualising
"""


import os.path
from typing import Dict, List, Mapping, Optional, Sequence, Tuple, Union

import numpy as np
import torch
from detectron2.layers import batched_nms
from detectron2.structures import BoxMode

from ..datapoint.annotation import ImageAnnotation
from ..datapoint.image import Image
from ..mapper.maputils import curry
from ..utils.detection_types import JsonDict
from ..utils.file_utils import wandb_available
from ..utils.settings import ObjectTypes, TypeOrStr

if wandb_available():
    from wandb import Classes
    from wandb import Image as Wbimage


@curry
def image_to_d2_frcnn_training(
    dp: Image,
    add_mask: bool = False,
    category_names: Optional[Union[str, ObjectTypes, Sequence[Union[str, ObjectTypes]]]] = None,
) -> Optional[JsonDict]:
    """
    Maps an image to a standard dataset dict as described in
    <https://detectron2.readthedocs.io/en/latest/tutorials/datasets.html>. It further checks if the image is physically
    available, for otherwise the annotation will be filtered.
    Note, that the returned dict will not suffice for training as gt for RPN and anchors still need to be created.

    :param dp: Image
    :param add_mask: True is not implemented (yet).
    :param category_names: A list of category names for training a model. Pass nothing to train with all annotations
    :return: Dict with 'image', 'width', 'height', 'image_id', 'annotations' where 'annotations' is a list of dict
             with 'bbox_mode' (D2 internal bounding box description), 'bbox' and 'category_id'.
    """
    if not os.path.isfile(dp.location) and dp.image is None:
        return None

    output: JsonDict = {"file_name": str(dp.location)}

    if dp.image is not None:
        output["image"] = dp.image.astype("float32")
    output["width"] = dp.width
    output["height"] = dp.height
    output["image_id"] = dp.image_id

    anns = dp.get_annotation(category_names=category_names)

    if not anns:
        return None

    annotations = []

    for ann in anns:
        if ann.image is not None:
            box = ann.image.get_embedding(dp.image_id)
        else:
            box = ann.bounding_box
        if box is None:
            raise ValueError("BoundingBox cannot be None")
        mapped_ann: Dict[str, Union[str, int, List[float]]] = {
            "bbox_mode": BoxMode.XYXY_ABS,
            "bbox": box.to_list(mode="xyxy"),
            "category_id": int(ann.category_id) - 1,
        }
        annotations.append(mapped_ann)

        if add_mask:
            raise NotImplementedError

    output["annotations"] = annotations

    return output


def pt_nms_image_annotations(
    anns: Sequence[ImageAnnotation], threshold: float, image_id: Optional[str] = None
) -> Sequence[str]:
    """
    Processing given image annotations through NMS. This is useful, if you want to supress some specific image
    annotation, e.g. given by name or returned through different predictors. This is the pt version, for tf check
    `mapper.tpstruct`

    :param anns: A sequence of ImageAnnotations. All annotations will be treated as if they belong to one category
    :param threshold: NMS threshold
    :param image_id: id in order to get the embedding bounding box

    :return: A list of annotation_ids that belong to the given input sequence and that survive the NMS process
    """
    if len(anns) == 1:
        return [anns[0].annotation_id]
    if not anns:
        return []
    ann_ids = np.array([ann.annotation_id for ann in anns], dtype="object")
    if image_id:
        boxes = torch.tensor(
            [ann.image.get_embedding(image_id).to_list(mode="xyxy") for ann in anns if ann.image is not None]
        )
        # if we do not have image embeddings but pass an image_id
        if not boxes.shape[0]:
            boxes = torch.tensor(
                [ann.bounding_box.to_list(mode="xyxy") for ann in anns if ann.bounding_box is not None]
            )
    else:
        boxes = torch.tensor([ann.bounding_box.to_list(mode="xyxy") for ann in anns if ann.bounding_box is not None])
    scores = torch.tensor([ann.score for ann in anns])
    class_mask = torch.ones(len(boxes), dtype=torch.uint8)
    keep = batched_nms(boxes, scores, class_mask, threshold)
    ann_ids_keep = ann_ids[keep]
    if not isinstance(ann_ids_keep, str):
        return ann_ids_keep.tolist()
    return []


@curry
def to_wandb_image(dp: Image, categories: Mapping[str, TypeOrStr]) -> Tuple[str, "Wbimage"]:
    """
    Converting a deepdoctection image into a wandb image

    :param dp: deepdoctection image
    :param categories: dict of categories
    :return: a W&B image
    """
    if dp.image is None:
        raise ValueError("Cannot convert to W&B image type when Image.image is None")

    boxes = []
    anns = dp.get_annotation(category_names=list(categories.values()))
    class_labels = {int(key): val for key, val in categories.items()}

    class_set = Classes([{"name": val, "id": int(key)} for key, val in categories.items()])

    for ann in anns:
        bounding_box = ann.image.get_embedding(dp.image_id) if ann.image is not None else ann.bounding_box
        box = {
            "position": {"middle": bounding_box.center, "width": bounding_box.width, "height": bounding_box.height},
            "domain": "pixel",
            "class_id": int(ann.category_id),
            "box_caption": ann.category_name,
        }
        if ann.score:
            box["scores"] = {"acc": ann.score}
        boxes.append(box)

    predictions = {"predictions": {"box_data": boxes, "class_labels": class_labels}}

    return dp.image_id, Wbimage(dp.image[:, :, ::-1], mode="RGB", boxes=predictions, classes=class_set)



# -*- coding: utf-8 -*-
# File: test_hfdetr.py

# Copyright 2023 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Testing module extern.hfdetr
"""


from typing import Dict, Sequence
from unittest.mock import MagicMock, patch

from pytest import mark

from deepdoctection.extern.hfdetr import HFDetrDerivedDetector
from deepdoctection.utils.detection_types import ImageType
from deepdoctection.utils.file_utils import pytorch_available, transformers_available
from deepdoctection.utils.settings import ObjectTypes

if pytorch_available():
    import torch

if transformers_available():
    from transformers import BatchFeature


def get_mock_features() -> "BatchFeature":
    """
    returns batch features inputs
    """
    data = {"pixel_values": torch.ones(1, 20, 20), "pixel_mask": torch.ones(1, 3, 20, 20)}
    return BatchFeature(data=data)


def get_mock_predictions() -> Dict[str, "torch.Tensor"]:
    """
    return detr predictions
    """

    outputs = {
        "encoder_last_hidden_state": torch.ones(10, 20, 3),
        "last_hidden_state": torch.ones(1, 2, 256),
        "logits": torch.tensor(
            [
                [
                    [-12.0546, -10.9208, -5.9281, -7.7109, -6.0824, -7.6029, 4.9027],
                    [-12.7695, -9.4253, -4.9289, -3.3625, -6.5684, -6.4647, 4.4205],
                ]
            ]
        ),
        "pred_boxes": torch.tensor([[[0.4978, 0.6739, 0.8992, 0.2794], [0.4972, 0.4949, 0.9043, 0.3606]]]),
    }

    return outputs


def get_mock_post_process_features() -> Sequence[Dict[str, "torch.Tensor"]]:
    """
    return feature_extractor.post_process_object_detection
    """

    return [
        {
            "boxes": torch.Tensor([[20, 10, 10, 10], [30, 40, 50, 60]]),
            "labels": torch.Tensor([0, 2]).to(torch.uint8),
            "scores": torch.Tensor([0.4695, 0.9522]),
        }
    ]


class TestHFDetrDerivedDetector:
    """
    Test HFDetrDerivedDetector
    """

    @staticmethod
    @mark.requires_pt
    @patch("deepdoctection.extern.hfdetr.HFDetrDerivedDetector.set_model", MagicMock(return_value=MagicMock()))
    @patch("deepdoctection.extern.hfdetr.HFDetrDerivedDetector.set_pre_processor", MagicMock())
    @patch("deepdoctection.extern.hfdetr.PretrainedConfig.from_pretrained", MagicMock())
    def test_hf_detr_predicts_image(detr_categories: Dict[str, ObjectTypes], np_image: ImageType) -> None:
        """
        D2 FRCNN calls predict_image and post processes DetectionResult correctly, e.g. adding class names
        """

        # Arrange
        detr = HFDetrDerivedDetector(
            path_config_json="",
            path_weights="",
            path_feature_extractor_config_json="",
            categories=detr_categories,
            device="cpu",
        )
        detr.hf_detr_predictor = MagicMock(return_value=get_mock_predictions())
        detr.feature_extractor = MagicMock(return_value=get_mock_features())
        detr.feature_extractor.post_process_object_detection = MagicMock(return_value=get_mock_post_process_features())
        # Act
        results = detr.predict(np_image)

        # Assert
        assert len(results) == 2
        assert results[0].class_id == 3
        assert results[1].class_id == 1



# -*- coding: utf-8 -*-
# File: test_tessocr.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Testing module extern.tessocr
"""
from typing import List
from unittest.mock import MagicMock, patch

import pytest

from deepdoctection.extern.base import DetectionResult
from deepdoctection.extern.tessocr import TesseractOcrDetector, tesseract_line_to_detectresult
from deepdoctection.utils.detection_types import ImageType
from deepdoctection.utils.file_utils import TesseractNotFound
from tests.data import Annotations

from .data import WORD_RESULTS


@pytest.fixture(name="word_result_list_same_line")
def fixture_pdf_bytes_page_2() -> List[DetectionResult]:
    """
    fixture list of word results. Words are in the same line
    """
    return WORD_RESULTS


def get_mock_word_results(
    np_img: ImageType, supported_languages: str, text_lines: bool, config: str  # pylint: disable=W0613
) -> List[DetectionResult]:
    """
    Returns WordResults attr: word_results_list
    """
    return Annotations().get_word_detect_results()


class TestTesseractOcrDetector:
    """
    Test TesseractOcrDetector
    """

    @staticmethod
    @pytest.mark.basic
    @patch("deepdoctection.utils.subprocess.check_output", MagicMock(side_effect=OSError))
    def test_tesseract_ocr_raises_tesseract_not_found_error_when_dependencies_not_satisfied(
        path_to_tesseract_yaml: str,
    ) -> None:
        """
        This tests shows that the dependency implementation of the base class and is representative for
        the dependency logic of all derived classes.
        """

        # Act and Assert
        with pytest.raises(TesseractNotFound):
            TesseractOcrDetector(path_yaml=path_to_tesseract_yaml)

    @staticmethod
    @pytest.mark.basic
    @patch("deepdoctection.utils.file_utils.get_tesseract_version", MagicMock(return_value=3.15))
    @patch("deepdoctection.extern.tessocr.predict_text", MagicMock(side_effect=get_mock_word_results))
    def test_tesseract_ocr_predicts_image(path_to_tesseract_yaml: str, np_image: ImageType) -> None:
        """
        Detector calls predict_text
        """

        # Arrange
        tess = TesseractOcrDetector(path_yaml=path_to_tesseract_yaml)

        # Act
        results = tess.predict(np_image)

        # Assert
        assert len(results) == 2


@pytest.mark.basic
def test_line_detect_result_returns_line(word_result_list_same_line: List[DetectionResult]) -> None:
    """
    Testing tesseract_line_to_detectresult generates Line DetectionResult
    """

    # Act
    detect_result_list = tesseract_line_to_detectresult(word_result_list_same_line)

    # Assert
    assert len(detect_result_list) == 3
    line_detect_result = detect_result_list[2]
    assert line_detect_result.box == [10.0, 10.0, 38.0, 24.0]
    assert line_detect_result.class_id == 2
    assert line_detect_result.text == "foo bak"



# -*- coding: utf-8 -*-
# File: IIITAR13K.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module for IIITar13K dataset. Install the dataset following the folder structure

    iiitar13K
    ├── test_images
    │ ├── ar_alphabet_2004_eng_38.jpg
    │ ├── ar_alphabet_2004_eng_39.jpg
    ├── test_xml
    │ ├── ar_alphabet_2004_eng_38.xml
    │ ├── ar_alphabet_2004_eng_39.xml
    ├── training_images
    │ ├── ...
    ├── training_xml
    │ ├── ...
    ├── validation_images
    │ ├── ...
    ├── validation_xml
    │ ├── ...

"""

import os
from typing import Mapping, Union

from ...dataflow import DataFlow, MapData, SerializerFiles
from ...datasets.info import DatasetInfo
from ...mapper.maputils import curry
from ...mapper.misc import xml_to_dict
from ...mapper.pascalstruct import pascal_voc_dict_to_image
from ...utils.detection_types import JsonDict
from ...utils.file_utils import lxml_available
from ...utils.settings import DatasetType, LayoutType
from ...utils.systools import get_package_path
from ..base import _BuiltInDataset
from ..dataflow_builder import DataFlowBaseBuilder
from ..info import DatasetCategories
from ..registry import dataset_registry

if lxml_available():
    from lxml import etree

_NAME = "iiitar13k"

_DESCRIPTION = (
    "[excerpt from Ajoy Mondal et. all. IIIT-AR-13K: A New Dataset for Graphical Object Detection in \n"
    "Documents] ...This dataset, IIIT-AR-13K, is created by manually annotating the bounding boxes of \n"
    "graphical or page objects in publicly available annual reports. This dataset contains a total of 13K \n"
    "annotated page images with objects in five different popular categories — table, figure, natural \n"
    "image, logo, and signature. This is the largest manually annotated dataset for graphical object \n"
    "detection. Annual reports created in multiple languages for several years from various companies \n"
    "bring high diversity into this dataset."
)

_LICENSE = "NN"

_URL = "http://cvit.iiit.ac.in/usodi/iiitar13k.php"

_SPLITS: Mapping[str, str] = {"train": "training_images", "val": "validation_images", "test": "test_images"}
_TYPE = DatasetType.object_detection
_LOCATION = "iiitar13k"
_ANNOTATION_FILES: Mapping[str, str] = {
    "train": "training_xml",
    "val": "validation_xml",
    "test": "test_xml",
}

_INIT_CATEGORIES = [LayoutType.table, LayoutType.logo, LayoutType.figure, LayoutType.signature]


@dataset_registry.register("iiitar13k")
class IIITar13K(_BuiltInDataset):
    """
    IIITar13K
    """

    _name = _NAME

    @classmethod
    def _info(cls) -> DatasetInfo:
        return DatasetInfo(name=_NAME, description=_DESCRIPTION, license=_LICENSE, url=_URL, splits=_SPLITS, type=_TYPE)

    def _categories(self) -> DatasetCategories:
        return DatasetCategories(init_categories=_INIT_CATEGORIES)

    def _builder(self) -> "IIITar13KBuilder":
        return IIITar13KBuilder(location=_LOCATION, annotation_files=_ANNOTATION_FILES)


class IIITar13KBuilder(DataFlowBaseBuilder):
    """
    IIITar13K dataflow builder
    """

    def build(self, **kwargs: Union[str, int]) -> DataFlow:
        """
        Returns a dataflow from which you can stream datapoints of images. The following arguments affect the return
        values of the dataflow:

        `split:` Split of the dataset. Can be `train`,`val` or `test`. Default: `val`

        `max_datapoints:` Will stop iterating after max_datapoints. Default: `None`

        `load_image:` Will load the image for each datapoint.  Default: `False`

        `fake_score:` Will add a fake score so that annotations look like predictions

        :return: dataflow
        """

        split = str(kwargs.get("split", "val"))
        load_image = kwargs.get("load_image", False)
        max_datapoints = kwargs.get("max_datapoints")
        fake_score = kwargs.get("fake_score", False)

        if max_datapoints is not None:
            max_datapoints = int(max_datapoints)

        # Load
        path_ann_files = self.get_workdir() / self.get_annotation_file(split)

        df = SerializerFiles.load(path_ann_files, ".xml", max_datapoints)
        utf8_parser = etree.XMLParser(encoding="utf-8")

        @curry
        def load_xml(path_ann: str, utf8_parser: etree.XMLParser) -> JsonDict:
            with open(path_ann, "r", encoding="utf-8") as xml_file:
                root = etree.fromstring(xml_file.read().encode("utf_8"), parser=utf8_parser)
            return {"file_name": path_ann, "xml": root}

        df = MapData(df, load_xml(utf8_parser))  # pylint: disable=E1120

        with open(
            os.path.join(get_package_path(), "deepdoctection/datasets/instances/xsl/pascal_voc.xsl"),
            "r",
            encoding="utf-8",
        ) as xsl_file:
            xslt_file = xsl_file.read().encode("utf-8")
        xml_obj = etree.XML(xslt_file, parser=etree.XMLParser(encoding="utf-8"))
        xslt_obj = etree.XSLT(xml_obj)

        df = MapData(df, xml_to_dict(xslt_obj))  # pylint: disable = E1120

        def _map_file_name(dp: JsonDict) -> JsonDict:
            dp["json"]["filename"] = dp["file_name"]
            return dp["json"]

        df = MapData(df, _map_file_name)
        df = MapData(
            df,
            pascal_voc_dict_to_image(  # pylint: disable = E1120
                self.categories.get_categories(init=True, name_as_key=True),
                load_image,
                filter_empty_image=True,
                fake_score=fake_score,
                category_name_mapping={
                    "natural_image": LayoutType.figure,
                    "figure": LayoutType.figure,
                    "logo": LayoutType.logo,
                    "signature": LayoutType.signature,
                    "table": LayoutType.table,
                },
            ),
        )

        return df



# -*- coding: utf-8 -*-
# File: test_utils.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Testing the module mapper.utils
"""
from typing import Dict, List, Union
from unittest.mock import MagicMock

import pytest
from pytest import mark

from deepdoctection.mapper import DefaultMapper, LabelSummarizer
from deepdoctection.utils.settings import ObjectTypes


class TestDefaultMapper:
    """
    Testing Class methods of DefaultMapper
    """

    @staticmethod
    @mark.basic
    def test_func_is_called_with_default_arguments() -> None:
        """
        Function in DefaultMapper is called with first argument of default_mapper  __call__ argument and all other
        arguments from DefaultMapper attributes.
        """

        # Arrange
        test_mapper = DefaultMapper(MagicMock(), "foo", "bak", baz="foo_bak")

        # Act
        test_mapper(dp="input")

        # Assert
        test_mapper.func.assert_called_with("input", "foo", "bak", baz="foo_bak")  # type: ignore


class TestLabelSummarizer:
    """
    Testing Class methods of LabelSummarizer
    """

    @staticmethod
    @mark.basic
    @pytest.mark.parametrize(
        "categories, cat_ids, summary",
        [
            ({"1": "FOO", "2": "BAK", "3": "BAZ"}, ["1", "3", "2", "2", "3"], {"1": 1, "2": 2, "3": 2}),
            ({"1": "FOO", "2": "BAK"}, ["1", "2", ["1", "1", "2"], "1", "2", ["1", "1"]], {"1": 6, "2": 3}),
            ({"1": "FOO", "2": "BAK", "3": "BAZ"}, [1, 3, 2, 2, "3", 1, 1, 1, 1, "1"], {"1": 6, "2": 2, "3": 2}),
        ],
    )
    def test_categories_are_correctly_summarized(
        categories: Dict[str, ObjectTypes], cat_ids: List[Union[List[Union[str, int]], str]], summary: Dict[str, int]
    ) -> None:
        """
        Testing Summarizer input with various dumped category id representations.
        """
        # Arrange
        summarizer = LabelSummarizer(categories)

        # Act
        for element in cat_ids:
            summarizer.dump(element)

        # Assert
        assert summarizer.get_summary() == summary



# -*- coding: utf-8 -*-
# File: model_rpn.py

# Copyright (c) Tensorpack Contributors
# Licensed under the Apache License, Version 2.0 (the "License")

"""
This file is modified from
<https://github.com/tensorpack/tensorpack/blob/master/examples/FasterRCNN/modeling/model_rpn.py>
"""

import numpy as np

# pylint: disable=import-error
import tensorflow as tf
from tensorpack import tfv1
from tensorpack.models import Conv2D, layer_register
from tensorpack.tfutils.argscope import argscope
from tensorpack.tfutils.scope_utils import auto_reuse_variable_scope, under_name_scope
from tensorpack.tfutils.summary import add_moving_summary
from tensorpack.utils.argtools import memoized

from .model_box import clip_boxes

# pylint: enable=import-error


@layer_register(log_shape=True)
@auto_reuse_variable_scope
def rpn_head(feature_map, channel, num_anchors):
    """
    RPN head

    :return: label_logits: fHxfWxNA
             box_logits: fHxfWxNAx4
    """

    with argscope(Conv2D, data_format="channels_first", kernel_initializer=tf.random_normal_initializer(stddev=0.01)):
        hidden = Conv2D("conv0", feature_map, channel, 3, activation=tf.nn.relu)

        label_logits = Conv2D("class", hidden, num_anchors, 1)
        box_logits = Conv2D("box", hidden, 4 * num_anchors, 1)
        # 1, NA(*4), im/16, im/16 (NCHW)

        label_logits = tf.transpose(label_logits, [0, 2, 3, 1])  # 1xfHxfWxNA
        label_logits = tf.squeeze(label_logits, 0)  # fHxfWxNA

        shp = tf.shape(box_logits)  # 1x(NAx4)xfHxfW
        box_logits = tf.transpose(box_logits, [0, 2, 3, 1])  # 1xfHxfWx(NAx4)
        box_logits = tf.reshape(box_logits, tf.stack([shp[2], shp[3], num_anchors, 4]))  # fHxfWxNAx4
    return label_logits, box_logits


@under_name_scope()
def rpn_losses(anchor_labels, anchor_boxes, label_logits, box_logits, rpn_batch_per_im):
    """
    RPN losses

    :param anchor_labels: fHxfWxNA
    :param anchor_boxes: fHxfWxNAx4, encoded
    :param label_logits: fHxfWxNA
    :param box_logits: fHxfWxNAx4
    :param rpn_batch_per_im: RPN batch per image

    :return: label_loss, box_loss
    """

    with tf.device("/cpu:0"):
        valid_mask = tf.stop_gradient(tf.not_equal(anchor_labels, -1))
        pos_mask = tf.stop_gradient(tf.equal(anchor_labels, 1))
        nr_valid = tf.stop_gradient(tfv1.count_nonzero(valid_mask, dtype=tf.int32), name="num_valid_anchor")
        nr_pos = tf.identity(tfv1.count_nonzero(pos_mask, dtype=tf.int32), name="num_pos_anchor")
        # nr_pos is guaranteed >0 in C4. But in FPN. even nr_valid could be 0.

        valid_anchor_labels = tf.boolean_mask(anchor_labels, valid_mask)
    valid_label_logits = tf.boolean_mask(label_logits, valid_mask)

    with tf.name_scope("label_metrics"):
        valid_label_prob = tf.nn.sigmoid(valid_label_logits)
        summaries = []
        with tf.device("/cpu:0"):
            for thresh in [0.5, 0.2, 0.1]:
                valid_prediction = tf.cast(valid_label_prob > thresh, tf.int32)
                nr_pos_prediction = tf.reduce_sum(valid_prediction, name="num_pos_prediction")
                pos_prediction_corr = tfv1.count_nonzero(
                    tf.logical_and(valid_label_prob > thresh, tf.equal(valid_prediction, valid_anchor_labels)),
                    dtype=tf.int32,
                )
                placeholder = 0.5  # A small value will make summaries appear lower.
                recall = tf.cast(tf.truediv(pos_prediction_corr, nr_pos), tf.float32)
                recall = tf.where(tf.equal(nr_pos, 0), placeholder, recall, name=f"recall_th{thresh}")
                precision = tf.cast(tf.truediv(pos_prediction_corr, nr_pos_prediction), tf.float32)
                precision = tf.where(
                    tf.equal(nr_pos_prediction, 0), placeholder, precision, name=f"precision_th{thresh}"
                )
                summaries.extend([precision, recall])
        add_moving_summary(*summaries)

    # Per-level loss summaries in FPN may appear lower due to the use of a small placeholder.
    # But the total RPN loss will be fine.  TODO make the summary op smarter
    placeholder = 0.0
    label_loss = tf.nn.sigmoid_cross_entropy_with_logits(
        labels=tf.cast(valid_anchor_labels, tf.float32), logits=valid_label_logits
    )
    label_loss = tf.reduce_sum(label_loss) * (1.0 / rpn_batch_per_im)
    label_loss = tf.where(tf.equal(nr_valid, 0), placeholder, label_loss, name="label_loss")

    pos_anchor_boxes = tf.boolean_mask(anchor_boxes, pos_mask)
    pos_box_logits = tf.boolean_mask(box_logits, pos_mask)
    delta = 1.0 / 9
    box_loss = (
        tfv1.losses.huber_loss(pos_anchor_boxes, pos_box_logits, delta=delta, reduction=tfv1.losses.Reduction.SUM)
        / delta
    )
    box_loss = box_loss * (1.0 / rpn_batch_per_im)
    box_loss = tf.where(tf.equal(nr_pos, 0), placeholder, box_loss, name="box_loss")

    add_moving_summary(label_loss, box_loss, nr_valid, nr_pos)
    return [label_loss, box_loss]


@under_name_scope()
def generate_rpn_proposals(
    boxes, scores, img_shape, rpn_min_size, rpn_proposal_nms_thres, pre_nms_top_k, post_nms_top_k=None
):
    """
    Sample RPN proposals by the following steps:
    1. Pick top k1 by scores
    2. NMS them
    3. Pick top k2 by scores. Default k2 == k1, i.e. does not filter the NMS output.

    :param post_nms_top_k: (int): See above.
    :param pre_nms_top_k: (int): See above.
    :param img_shape: [h, w]
    :param scores: n float, the logits
    :param boxes: nx4 float dtype, the proposal boxes. Decoded to float box already

    :return: boxes: kx4 float
            scores: k logits

    """
    assert boxes.shape.ndims == 2, boxes.shape
    if post_nms_top_k is None:
        post_nms_top_k = pre_nms_top_k

    top_k = tf.minimum(pre_nms_top_k, tf.size(scores))
    top_k_scores, top_k_indices = tf.nn.top_k(scores, k=top_k, sorted=False)
    top_k_boxes = tf.gather(boxes, top_k_indices)
    top_k_boxes = clip_boxes(top_k_boxes, img_shape)

    if rpn_min_size > 0:
        top_k_boxes_x1y1x2y2 = tf.reshape(top_k_boxes, (-1, 2, 2))
        top_k_boxes_x1y1, top_k_boxes_x2y2 = tf.split(top_k_boxes_x1y1x2y2, 2, axis=1)
        # nx1x2 each
        wbhb = tf.squeeze(top_k_boxes_x2y2 - top_k_boxes_x1y1, axis=1)
        valid = tf.reduce_all(wbhb > rpn_min_size, axis=1)  # n,
        top_k_valid_boxes = tf.boolean_mask(top_k_boxes, valid)
        top_k_valid_scores = tf.boolean_mask(top_k_scores, valid)
    else:
        top_k_valid_boxes = top_k_boxes
        top_k_valid_scores = top_k_scores

    nms_indices = tf.image.non_max_suppression(
        top_k_valid_boxes, top_k_valid_scores, max_output_size=post_nms_top_k, iou_threshold=rpn_proposal_nms_thres
    )

    proposal_boxes = tf.gather(top_k_valid_boxes, nms_indices)
    proposal_scores = tf.gather(top_k_valid_scores, nms_indices)
    tf.sigmoid(proposal_scores, name="probs")  # for visualization
    return tf.stop_gradient(proposal_boxes, name="boxes"), tf.stop_gradient(proposal_scores, name="scores")


@memoized
def get_all_anchors(*, stride, sizes, ratios, max_size):
    """
    Get all anchors in the largest possible image, shifted, float box

    :param stride: the stride of anchors.
    :param sizes: the sizes (sqrt area) of anchors
    :param ratios: the aspect ratios of anchors
    :param max_size: maximum size of input image

    :return: SxSxNUM_ANCHOR x 4, where S == ceil(MAX_SIZE/STRIDE), float box

    The layout in the NUM_ANCHOR dim is NUM_RATIO x NUM_SIZE.
    """
    # Generates a NAx4 matrix of anchor boxes in (x1, y1, x2, y2) format. Anchors
    # are centered on 0, have sqrt areas equal to the specified sizes, and aspect ratios as given.
    anchors = []
    for size in sizes:
        for ratio in ratios:
            w = np.sqrt(size * size / ratio)
            h = ratio * w
            anchors.append([-w, -h, w, h])
    cell_anchors = np.asarray(anchors) * 0.5

    field_size = int(np.ceil(max_size / stride))
    shifts = (np.arange(0, field_size) * stride).astype("float32")
    shift_x, shift_y = np.meshgrid(shifts, shifts)
    shift_x = shift_x.flatten()
    shift_y = shift_y.flatten()
    shifts = np.vstack((shift_x, shift_y, shift_x, shift_y)).transpose()
    # Kx4, K = field_size * field_size
    K = shifts.shape[0]  # pylint: disable =C0103

    A = cell_anchors.shape[0]  # pylint: disable =C0103
    field_of_anchors = cell_anchors.reshape((1, A, 4)) + shifts.reshape((1, K, 4)).transpose((1, 0, 2))
    field_of_anchors = field_of_anchors.reshape((field_size, field_size, A, 4))
    # FSxFSxAx4
    # Many rounding happens inside the anchor code anyway
    # assert np.all(field_of_anchors == field_of_anchors.astype('int32'))
    field_of_anchors = field_of_anchors.astype("float32")
    return field_of_anchors



# -*- coding: utf-8 -*-
# File: model_box.py

# Copyright (c) Tensorpack Contributors
# Licensed under the Apache License, Version 2.0 (the "License")

"""
This file is modified from
<https://github.com/tensorpack/tensorpack/blob/master/examples/FasterRCNN/modeling/model_box.py>
"""
from collections import namedtuple

import numpy as np

# pylint: disable=import-error
import tensorflow as tf
from tensorpack.tfutils.scope_utils import under_name_scope

# pylint: enable=import-error


@under_name_scope()
def clip_boxes(boxes, window, name=None):
    """
    clip boxes

    :param boxes: nx4, xyxy
    :param window: [h, w]
    :param name: (str)
    :return:
    """
    boxes = tf.maximum(boxes, 0.0)
    mat = tf.tile(tf.reverse(window, [0]), [2])  # (4,)
    boxes = tf.minimum(boxes, tf.cast(mat, tf.float32), name=name)
    return boxes


@under_name_scope()
def decode_bbox_target(box_predictions, anchors, preproc_max_size):
    """
    Decode bbox target

    :param box_predictions: (..., 4), logits
    :param anchors: (..., 4), float box. Must have the same shape
    :param preproc_max_size: int
    :return: (..., 4), float32. With the same shape.
    """

    orig_shape = tf.shape(anchors)
    box_pred_txtytwth = tf.reshape(box_predictions, (-1, 2, 2))
    box_pred_txty, box_pred_twth = tf.split(box_pred_txtytwth, 2, axis=1)
    # each is (...)x1x2
    anchors_x1y1x2y2 = tf.reshape(anchors, (-1, 2, 2))
    anchors_x1y1, anchors_x2y2 = tf.split(anchors_x1y1x2y2, 2, axis=1)

    waha = anchors_x2y2 - anchors_x1y1
    xaya = (anchors_x2y2 + anchors_x1y1) * 0.5

    clip = np.log(preproc_max_size / 16.0)
    wbhb = tf.exp(tf.minimum(box_pred_twth, clip)) * waha
    xbyb = box_pred_txty * waha + xaya
    x1y1 = xbyb - wbhb * 0.5
    x2y2 = xbyb + wbhb * 0.5  # (...)x1x2
    out = tf.concat([x1y1, x2y2], axis=-2)
    return tf.reshape(out, orig_shape)


@under_name_scope()
def encode_bbox_target(boxes, anchors):
    """
    Encode bbox target

    :param boxes: (..., 4), float32
    :param anchors: (..., 4), float32
    :return: (..., 4), float32 with the same shape.
    """

    anchors_x1y1x2y2 = tf.reshape(anchors, (-1, 2, 2))
    anchors_x1y1, anchors_x2y2 = tf.split(anchors_x1y1x2y2, 2, axis=1)
    waha = anchors_x2y2 - anchors_x1y1
    xaya = (anchors_x2y2 + anchors_x1y1) * 0.5

    boxes_x1y1x2y2 = tf.reshape(boxes, (-1, 2, 2))
    boxes_x1y1, boxes_x2y2 = tf.split(boxes_x1y1x2y2, 2, axis=1)
    wbhb = boxes_x2y2 - boxes_x1y1
    xbyb = (boxes_x2y2 + boxes_x1y1) * 0.5

    # Note that here not all boxes are valid. Some may be zero
    txty = (xbyb - xaya) / waha
    twth = tf.math.log(wbhb / waha)  # may contain -inf for invalid boxes
    encoded = tf.concat([txty, twth], axis=1)  # (-1x2x2)
    return tf.reshape(encoded, tf.shape(boxes))


@under_name_scope()
def crop_and_resize(image, boxes, box_ind, crop_size, pad_border=True):
    """
    Crop and resize

    :param image: NCHW
    :param boxes: nx4, x1y1x2y2
    :param box_ind: (n,)
    :param crop_size: (int)
    :param pad_border:  bool
    :return:   n,C,size,size
    """

    assert isinstance(crop_size, int), crop_size
    boxes = tf.stop_gradient(boxes)

    # TF's crop_and_resize produces zeros on border
    if pad_border:
        # this can be quite slow
        image = tf.pad(image, [[0, 0], [0, 0], [1, 1], [1, 1]], mode="SYMMETRIC")
        boxes += 1

    @under_name_scope()
    def transform_fpcoor_for_tf(boxes, image_shape, crop_shape):
        """
        The way tf.image.crop_and_resize works (with normalized box):
        Initial point (the value of output[0]): x0_box * (W_img - 1)
        Spacing: w_box * (W_img - 1) / (W_crop - 1)
        Use the above grid to bi linear sample.

        However, what we want is (with fpcoor box):
        Spacing: w_box / W_crop
        Initial point: x0_box + spacing/2 - 0.5
        (-0.5 because bi linear sample (in my definition) assumes floating point coordinate
         (0.0, 0.0) is the same as pixel value (0, 0))

        This function transform fpcoor boxes to a format to be used by tf.image.crop_and_resize

        :param boxes:  nx4, x1y1x2y2
        :param image_shape: shape
        :param crop_shape: crop shape
        :return:  y1x1y2x2
        """

        x_0, y_0, x_1, y_1 = tf.split(boxes, 4, axis=1)

        spacing_w = (x_1 - x_0) / tf.cast(crop_shape[1], tf.float32)
        spacing_h = (y_1 - y_0) / tf.cast(crop_shape[0], tf.float32)

        imshape = [tf.cast(image_shape[0] - 1, tf.float32), tf.cast(image_shape[1] - 1, tf.float32)]
        nx0 = (x_0 + spacing_w / 2 - 0.5) / imshape[1]
        ny0 = (y_0 + spacing_h / 2 - 0.5) / imshape[0]

        n_w = spacing_w * tf.cast(crop_shape[1] - 1, tf.float32) / imshape[1]
        n_h = spacing_h * tf.cast(crop_shape[0] - 1, tf.float32) / imshape[0]

        return tf.concat([ny0, nx0, ny0 + n_h, nx0 + n_w], axis=1)

    image_shape = tf.shape(image)[2:]

    boxes = transform_fpcoor_for_tf(boxes, image_shape, [crop_size, crop_size])
    image = tf.transpose(image, [0, 2, 3, 1])  # nhwc
    ret = tf.image.crop_and_resize(image, boxes, tf.cast(box_ind, tf.int32), crop_size=[crop_size, crop_size])
    ret = tf.transpose(ret, [0, 3, 1, 2])  # ncss
    return ret


@under_name_scope()
def roi_align(feature_map, boxes, resolution):
    """
    Roi align

    :param feature_map: 1xCxHxW
    :param boxes:  Nx4 float box
    :param resolution: output spatial resolution
    :return: Roi aligned tf.Tensor
    """

    # sample 4 locations per roi bin
    ret = crop_and_resize(feature_map, boxes, tf.zeros(tf.shape(boxes)[0], dtype=tf.int32), resolution * 2)
    try:
        avg_pool = tf.nn.avg_pool2d
    except AttributeError:
        avg_pool = tf.nn.avg_pool
    ret = avg_pool(ret, [1, 1, 2, 2], [1, 1, 2, 2], padding="SAME", data_format="NCHW")
    return ret


class RPNAnchors(namedtuple("_RPNAnchors", ["boxes", "gt_labels", "gt_boxes"])):
    """
    boxes (FS x FS x NA x 4): The anchor boxes.
    gt_labels (FS x FS x NA):
    gt_boxes (FS x FS x NA x 4): Ground-truth boxes corresponding to each anchor.
    """

    def encoded_gt_boxes(self):
        """
        encoded ground truth boxes
        """
        return encode_bbox_target(self.gt_boxes, self.boxes)

    def decode_logits(self, logits, preproc_max_size):
        """
        Decode logits

        :param logits: logits
        :param preproc_max_size: preprocess to max size
        """
        return decode_bbox_target(logits, self.boxes, preproc_max_size)

    @under_name_scope()
    def narrow_to(self, featuremap):
        """
        Slice anchors to the spatial size of this feature map.
        """
        shape2d = tf.shape(featuremap)[2:]  # h,w
        slice3d = tf.concat([shape2d, [-1]], axis=0)
        slice4d = tf.concat([shape2d, [-1, -1]], axis=0)
        boxes = tf.slice(self.boxes, [0, 0, 0, 0], slice4d)
        gt_labels = tf.slice(self.gt_labels, [0, 0, 0], slice3d)
        gt_boxes = tf.slice(self.gt_boxes, [0, 0, 0, 0], slice4d)
        return RPNAnchors(boxes, gt_labels, gt_boxes)



# -*- coding: utf-8 -*-
# File: test_custom_serialize.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Testing the module dataflow.custom_serialize
"""
import os

from pytest import mark

from deepdoctection.dataflow import (
    CustomDataFromList,
    SerializerCoco,
    SerializerFiles,
    SerializerJsonlines,
    SerializerPdfDoc,
    SerializerTabsepFiles,
)

from ..test_utils import collect_datapoint_from_dataflow, get_test_path


class TestSerializerJsonlines:
    """
    Testing SerializerJsonlines loading and saving functions
    """

    @staticmethod
    @mark.basic
    def test_loading() -> None:
        """
        Testing the loading of a .jsonl file
        """

        # Arrange
        test_file = os.path.join(get_test_path(), "test_file.jsonl")

        # Act
        df = SerializerJsonlines.load(test_file)
        output = collect_datapoint_from_dataflow(df=df)

        # Assert
        assert len(output) == 3

    @staticmethod
    @mark.basic
    def test_saving() -> None:
        """
        Testing the saving of a .jsonl file
        """

        # Arrange
        for_saving = [{"foo": "bak"}, {"fok": "bah"}]
        df = CustomDataFromList(for_saving)

        file_name = "test_file_saved.jsonl"
        test_file = os.path.join(get_test_path(), file_name)

        # Act
        SerializerJsonlines.save(df, get_test_path(), file_name)

        # Assert
        assert os.path.isfile(test_file)

        # Clean-up
        os.remove(test_file)


class TestSerializerFiles:
    """
    Testing SerializerFiles loading function
    """

    @staticmethod
    @mark.basic
    def test_loading() -> None:
        """
        Test the loading of elements of a .jsonl file
        """

        # Arrange and Act
        df = SerializerFiles.load(get_test_path(), file_type=".jsonl")
        output = collect_datapoint_from_dataflow(df=df)

        # Assert
        assert len(output) == 4


class TestSerializerTabsepFiles:
    """
    Testing SerializerTabsepFiles loading function
    """

    @staticmethod
    @mark.basic
    def test_loading() -> None:
        """
        Test the loading of elements of a .txt file
        """

        # Arrange and Act
        df = SerializerTabsepFiles.load(get_test_path() / "test_file.txt")
        output = collect_datapoint_from_dataflow(df=df)

        # Assert
        assert len(output) == 5


class TestSerializerCoco:
    """
    Testing SerializerCoco loading function
    """

    @staticmethod
    @mark.basic
    def test_loading() -> None:
        """
        Test the loading of a .json file
        """

        # Arrange
        test_file = os.path.join(get_test_path(), "test_file.json")

        # Act
        df = SerializerCoco.load(test_file)
        output = collect_datapoint_from_dataflow(df=df)

        # Assert
        assert len(output) == 20

        # Act
        df = SerializerCoco.load(test_file, max_datapoints=5)
        output = collect_datapoint_from_dataflow(df=df)

        # Assert
        assert len(output) == 5


class TestSerializerPdfDoc:
    """
    Testing SerializerPdfDoc loading function
    """

    @staticmethod
    @mark.basic
    def test_loading() -> None:
        """
        Test the loading of a .pdf file
        """

        # Arrange
        test_file = os.path.join(get_test_path(), "test_file.pdf")

        # Act
        df = SerializerPdfDoc.load(test_file)
        output = collect_datapoint_from_dataflow(df=df)
        first_image = output[0]

        # Assert
        assert len(output) == 2
        assert first_image["path"] == test_file
        assert first_image["file_name"] == "test_file_0.pdf"
        assert isinstance(first_image["pdf_bytes"], bytes)



# -*- coding: utf-8 -*-
# File: anngen.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module for datapoint populating helpers
"""
from dataclasses import asdict
from typing import Dict, List, Mapping, Optional, Union

import numpy as np

from ..datapoint.annotation import CategoryAnnotation, ContainerAnnotation, ImageAnnotation, SummaryAnnotation
from ..datapoint.box import BoundingBox, local_to_global_coords, rescale_coords
from ..datapoint.image import Image
from ..extern.base import DetectionResult
from ..mapper.maputils import MappingContextManager
from ..utils.settings import ObjectTypes, Relationships


class DatapointManager:
    """
    Class whose methods provide an API for manipulating image datapoints. This includes the creation and storage of
    annotations in the cache of the image but also in the annotations themselves.

    When the image is transferred, the annotations are stored in a cache dictionary so that access via the annotation ID
    can be performed efficiently.

    The manager is part of each `PipelineComponent`.
    """

    def __init__(self) -> None:
        self._datapoint: Optional[Image] = None
        self._cache_anns: Dict[str, ImageAnnotation] = {}
        self.datapoint_is_passed: bool = False
        self.category_id_mapping: Optional[Mapping[int, int]] = None

    @property
    def datapoint(self) -> Image:
        """
        datapoint
        """
        if self._datapoint is not None:
            return self._datapoint
        raise ValueError("no datapoint passed")

    @datapoint.setter
    def datapoint(self, dp: Image) -> None:
        """
        datapoint
        """
        self._datapoint = dp
        self._cache_anns = {ann.annotation_id: ann for ann in dp.get_annotation()}
        self.datapoint_is_passed = True

    def assert_datapoint_passed(self) -> None:
        """
        assert that datapoint is passed
        """
        assert self.datapoint_is_passed, "Pass datapoint to  DatapointManager before creating anns"

    def maybe_map_category_id(self, category_id: Union[str, int]) -> int:
        """
        Maps categories if a category id mapping is provided in `__init__`.

        :param category_id: category id via integer or string.
        :return: mapped category id
        """
        if self.category_id_mapping is None:
            return int(category_id)
        return self.category_id_mapping[int(category_id)]

    def set_category_id_mapping(self, category_id_mapping: Mapping[int, int]) -> None:
        """
        In many cases the category ids sent back from a model have to be modified. Pass a mapping from model
        category ids to target annotation category ids.

        :param category_id_mapping: A mapping of model category ids (sent from DetectionResult) to category ids (saved
                                    in annotations)
        """
        self.category_id_mapping = category_id_mapping

    def set_image_annotation(
        self,
        detect_result: DetectionResult,
        to_annotation_id: Optional[str] = None,
        to_image: bool = False,
        crop_image: bool = False,
        detect_result_max_width: Optional[float] = None,
        detect_result_max_height: Optional[float] = None,
    ) -> Optional[str]:
        """
        Creating an image annotation from a raw `DetectionResult` dataclass. Beside dumping the annotation to the
        `ImageAnnotation` cache you can also dump the annotation to the `image` of an annotation with given
        `annotation_id`. This is handy if, you know, you want to send the sub image to a subsequent pipeline component.

        Moreover, it is possible to generate an Image of the given raw annotation and store it in its `image`. The
        resulting image is given as a sub image of `self` defined by it bounding box coordinates. Use `crop_image`
        to explicitly store the sub image as numpy array.

        :param detect_result: A `DetectionResult` in general coming from ObjectDetector
        :param to_annotation_id: Will dump the created image annotation to `image` of the given annotation_id.
                                 Requires the to_annotation to have a not `None` image.
        :param to_image: If True will populate `image`.
        :param crop_image: Makes only sense if to_image=True and if a numpy array is stored in the original image.
                           Will generate `Image.image`.
        :param detect_result_max_width: If detect result has a different scaling scheme from the image it refers to,
                                        pass the max width possible so coords can be rescaled.
        :param detect_result_max_height: If detect result has a different scaling scheme from the image it refers to,
                                        pass the max height possible so coords can be rescaled.
        :return: the annotation_id of the generated image annotation
        """
        self.assert_datapoint_passed()
        assert detect_result.class_id, detect_result.class_id
        if not isinstance(detect_result.box, (list, np.ndarray)):
            raise TypeError(
                f"detect_result.box must be of type list or np.ndarray, " f"but is of type {(type(detect_result.box))}"
            )
        detect_result.class_id = self.maybe_map_category_id(detect_result.class_id)
        with MappingContextManager(
            dp_name=self.datapoint.file_name, filter_level="annotation", detect_result=asdict(detect_result)
        ) as annotation_context:
            box = BoundingBox(
                ulx=detect_result.box[0],
                uly=detect_result.box[1],
                lrx=detect_result.box[2],
                lry=detect_result.box[3],
                absolute_coords=detect_result.absolute_coords,
            )
            if detect_result_max_width and detect_result_max_height:
                box = rescale_coords(
                    box,
                    detect_result_max_width,
                    detect_result_max_height,
                    self.datapoint.width,
                    self.datapoint.height,
                )
            ann = ImageAnnotation(
                category_name=detect_result.class_name,
                bounding_box=box,
                category_id=str(detect_result.class_id),
                score=detect_result.score,
            )
            if to_annotation_id is not None:
                parent_ann = self._cache_anns[to_annotation_id]
                assert parent_ann.image, parent_ann.image
                parent_ann.image.dump(ann)
                parent_ann.image.image_ann_to_image(ann.annotation_id)
                ann_global_box = local_to_global_coords(
                    ann.bounding_box, parent_ann.image.get_embedding(self.datapoint.image_id)  # type: ignore
                )
                assert ann.image, ann.image
                ann.image.set_embedding(parent_ann.annotation_id, ann.bounding_box)
                ann.image.set_embedding(self.datapoint.image_id, ann_global_box)
                parent_ann.dump_relationship(Relationships.child, ann.annotation_id)

            self.datapoint.dump(ann)
            self._cache_anns[ann.annotation_id] = ann

            if to_image and to_annotation_id is None:
                self.datapoint.image_ann_to_image(annotation_id=ann.annotation_id, crop_image=crop_image)

        if annotation_context.context_error:
            return None
        return ann.annotation_id

    def set_category_annotation(
        self,
        category_name: ObjectTypes,
        category_id: Optional[Union[str, int]],
        sub_cat_key: ObjectTypes,
        annotation_id: str,
        score: Optional[float] = None,
    ) -> Optional[str]:
        """
        Create a category annotation and dump it as sub category to an already created annotation.

        :param category_name: category name
        :param category_id: category id
        :param sub_cat_key: the key to dump the created annotation to.
        :param annotation_id: id, of the parent annotation. Currently, this can only be an image annotation.
        :param score: Add a score.
        :return: the annotation_id of the generated category annotation
        """
        self.assert_datapoint_passed()
        with MappingContextManager(
            dp_name=self.datapoint.file_name,
            filter_level="annotation",
            category_annotation={
                "category_name": category_name.value,
                "sub_cat_key": sub_cat_key.value,
                "annotation_id": annotation_id,
            },
        ) as annotation_context:
            cat_ann = CategoryAnnotation(category_name=category_name, category_id=str(category_id), score=score)
            self._cache_anns[annotation_id].dump_sub_category(sub_cat_key, cat_ann)
        if annotation_context.context_error:
            return None
        return cat_ann.annotation_id

    def set_container_annotation(
        self,
        category_name: ObjectTypes,
        category_id: Optional[Union[str, int]],
        sub_cat_key: ObjectTypes,
        annotation_id: str,
        value: Union[str, List[str]],
        score: Optional[float] = None,
    ) -> Optional[str]:
        """
        Create a container annotation and dump it as sub category to an already created annotation.

        :param category_name: category name
        :param category_id: category id
        :param sub_cat_key: the key to dump the created annotation to.
        :param annotation_id: id, of the parent annotation. Currently, this can only be an image annotation.
        :param value: A value to store
        :param score: Add a score.
        :return: annotation_id of the generated container annotation
        """
        self.assert_datapoint_passed()
        with MappingContextManager(
            dp_name=self.datapoint.file_name,
            filter_level="annotation",
            container_annotation={
                "category_name": category_name.value,
                "sub_cat_key": sub_cat_key.value,
                "annotation_id": annotation_id,
                "value": str(value),
            },
        ) as annotation_context:
            cont_ann = ContainerAnnotation(
                category_name=category_name, category_id=str(category_id), value=value, score=score
            )
            self._cache_anns[annotation_id].dump_sub_category(sub_cat_key, cont_ann)
        if annotation_context.context_error:
            return None
        return cont_ann.annotation_id

    def set_summary_annotation(
        self,
        summary_key: ObjectTypes,
        summary_name: ObjectTypes,
        summary_number: int,
        summary_value: Optional[str] = None,
        summary_score: Optional[float] = None,
        annotation_id: Optional[str] = None,
    ) -> Optional[str]:
        """
        Creates a sub category of a summary annotation. If a summary of the given `annotation_id` does not exist, it
        will create a new one.
        :param summary_key: will store the category annotation as sub category
        :param summary_name: will create the summary name as category name
        :param summary_number: will store the value in category_id.
        :param summary_value: will create a ContainerAnnotation and store the corresponding value
        :param summary_score: will store the score
        :param annotation_id: id of the parent annotation. Note, that the parent annotation must have `image` to
        be not None.
        :return: `annotation_id` of the generated category annotation
        """
        self.assert_datapoint_passed()
        if annotation_id is not None:
            image = self._cache_anns[annotation_id].image
        else:
            image = self.datapoint
        assert image is not None, image
        if image.summary is None:
            image.summary = SummaryAnnotation()

        ann: Union[CategoryAnnotation, ContainerAnnotation]
        with MappingContextManager(
            dp_name=annotation_id,
            filter_level="annotation",
            summary_annotation={
                "summary_key": summary_key.value,
                "summary_name": summary_name.value,
                "summary_value": summary_value,
                "annotation_id": annotation_id,
            },
        ) as annotation_context:
            if summary_value:
                ann = ContainerAnnotation(
                    category_name=summary_name,
                    category_id=str(summary_number),
                    value=summary_value,
                    score=summary_score,
                )
            else:
                ann = CategoryAnnotation(
                    category_name=summary_name, category_id=str(summary_number), score=summary_score
                )
            image.summary.dump_sub_category(summary_key, ann, image.image_id)

        if annotation_context.context_error:
            return None
        return ann.annotation_id

    def deactivate_annotation(self, annotation_id: str) -> None:
        """
        Deactivate annotation by given annotation_id

        :param annotation_id: annotation_id
        """
        ann = self._cache_anns[annotation_id]
        ann.deactivate()
